/*
 *
 * (C) 2004-2005 Denora Team
 * Contact us at info@nomadirc.net
 *
 * Please read COPYING and README for furhter details.
 *
 * Based on the original code of Anope by Anope Team.
 * Based on the original code of Thales by Lucas.
 * 
 * $Id: channels.c 443 2005-08-01 19:51:49Z trystan $
 *
 */

#include "denora.h"

Channel *chanlist[1024];
StatsChannel *statschanlist[1024];
static Channel *current;
static StatsChannel *currentsc;
static int next_index;

/*************************************************************************/

void load_chan_db(void)
{
    DenoraDBFile *dbptr = scalloc(1, sizeof(DenoraDBFile));
    StatsChannel *ss = NULL;
    char *key, *value;
    int retval = 0;

    fill_db_ptr(dbptr, 0, CHAN_VERSION, s_StatServ, ChannelDB);
    SET_SEGV_LOCATION();

    /* let's remove existing temp files here, because we only load dbs on startup */
    remove(dbptr->temp_name);

    /* Open the db, fill the rest of dbptr and allocate memory for key and value */
    if (new_open_db_read(dbptr, &key, &value)) {
        SET_SEGV_LOCATION();
        free(dbptr);
        return;                 /* Bang, an error occurred */
    }

    while (1) {
        /* read a new entry and fill key and value with it -Certus */
        retval = new_read_db_entry(&key, &value, dbptr->fptr);

        if (retval == DB_READ_ERROR) {
            alog(LOG_NORMAL, langstr(ALOG_DB_ERROR), dbptr->filename);
            new_close_db(dbptr->fptr, &key, &value);
            SET_SEGV_LOCATION();
            free(dbptr);
            return;
        } else if (retval == DB_EOF_ERROR) {
            alog(LOG_EXTRADEBUG, langstr(ALOG_DEBUG_DB_OK),
                 dbptr->filename);
            new_close_db(dbptr->fptr, &key, &value);
            SET_SEGV_LOCATION();
            free(dbptr);
            return;
        } else if (retval == DB_READ_BLOCKEND) {        /* DB_READ_BLOCKEND */
            /* a channel has completely been read. put any checks in here! */
        } else {
            /* DB_READ_SUCCESS */

            if (!*value || !*key) {
                continue;
            }
            SET_SEGV_LOCATION();

            if (!stricmp(key, "name")) {
                ss = statschan_create(value);
            } else if (!stricmp(key, "kickcnt")) {
                ss->kickcount = atoi(value);
            } else if (!stricmp(key, "joincnt")) {
                ss->joincounter = atoi(value);
            } else if (!stricmp(key, "topiccnt")) {
                ss->topic_count = atoi(value);
            } else if (!stricmp(key, "maxusercnt")) {
                ss->maxusercount = atoi(value);
            } else if (!stricmp(key, "maxusertime")) {
                ss->maxusertime = atoi(value);
            }
        }                       /* else */
    }                           /* while */
}

/*************************************************************************/

void save_chan_db(void)
{
    DenoraDBFile *dbptr = scalloc(1, sizeof(DenoraDBFile));
    StatsChannel *ss;
    int i;

    fill_db_ptr(dbptr, 0, CHAN_VERSION, s_StatServ, ChannelDB);
    SET_SEGV_LOCATION();

    /* time to backup the old db */
    rename(ChannelDB, dbptr->temp_name);

    if (new_open_db_write(dbptr)) {
        rename(dbptr->temp_name, ChannelDB);
        free(dbptr);
        return;                 /* Bang, an error occurred */
    }
    SET_SEGV_LOCATION();
    for (i = 0; i < 1024; i++) {
        for (ss = statschanlist[i]; ss; ss = ss->next) {
            new_write_db_entry("name", dbptr, "%s", ss->name);
            new_write_db_entry("kickcnt", dbptr, "%d", ss->kickcount);
            new_write_db_entry("joincnt", dbptr, "%d", ss->joincounter);
            new_write_db_entry("topiccnt", dbptr, "%d", ss->topic_count);
            new_write_db_entry("maxusercnt", dbptr, "%d",
                               ss->maxusercount);
            new_write_db_entry("maxusertime", dbptr, "%ld",
                               (long int) ss->maxusertime);
            new_write_db_endofblock(dbptr);
        }
    }
    SET_SEGV_LOCATION();
    if (dbptr) {
        new_close_db(dbptr->fptr, NULL, NULL);  /* close file */
        remove(dbptr->temp_name);       /* saved successfully, no need to keep the old one */
        free(dbptr);            /* free the db struct */
    }
}

/*************************************************************************/

void sql_channel_ban(int type, char *chan, char *mask)
{
    SET_SEGV_LOCATION();

    if (LargeNet) {
        return;
    }

    chan = rdb_escape(chan);
    if (mask) {
        mask = rdb_escape(mask);
    }

    if (type == 1) {
        rdb_query
            ("INSERT IGNORE INTO %s (chan, bans) VALUES (\'%s\', \'%s\')",
             ChanBansTable, chan, mask);
    } else if (type == 2) {
        rdb_query("DELETE FROM %s WHERE chan=\'%s\' AND bans=\'%s\'",
                  ChanBansTable, chan, mask);
    } else {
        rdb_query("DELETE FROM %s WHERE chan=\'%s\'", ChanBansTable, chan);
    }
    SET_SEGV_LOCATION();
    denora_free(chan);
    if (mask) {
        denora_free(mask);
    }
}

/*************************************************************************/

void sql_channel_exception(int type, char *chan, char *mask)
{
    SET_SEGV_LOCATION();

    if (LargeNet) {
        return;
    }

    chan = rdb_escape(chan);
    if (mask) {
        mask = rdb_escape(mask);
    }

    if (type == 1) {
        rdb_query
            ("INSERT IGNORE INTO %s (chan, mask) VALUES (\'%s\', \'%s\')",
             ChanExceptTable, chan, mask);
    } else if (type == 2) {
        rdb_query("DELETE FROM %s WHERE chan=\'%s\' AND mask=\'%s\'",
                  ChanExceptTable, chan, mask);
    } else {
        rdb_query("DELETE FROM %s WHERE chan=\'%s\'", ChanExceptTable,
                  chan);
    }
    SET_SEGV_LOCATION();
    denora_free(chan);
    if (mask) {
        denora_free(mask);
    }
}

/*************************************************************************/

void sql_channel_invite(int type, char *chan, char *mask)
{
    SET_SEGV_LOCATION();

    if (LargeNet) {
        return;
    }

    chan = rdb_escape(chan);
    if (mask) {
        mask = rdb_escape(mask);
    }

    if (type == 1) {
        rdb_query
            ("INSERT IGNORE INTO %s (chan, mask) VALUES (\'%s\', \'%s\')",
             ChanInviteTable, chan, mask);
    } else if (type == 2) {
        rdb_query("DELETE FROM %s WHERE chan=\'%s\' AND mask=\'%s\'",
                  ChanInviteTable, chan, mask);
    } else {
        rdb_query("DELETE FROM %s WHERE chan=\'%s\'", ChanInviteTable,
                  chan);
    }
    SET_SEGV_LOCATION();
    denora_free(chan);
    if (mask) {
        denora_free(mask);
    }
}

/*************************************************************************/

/* adds modes to a given chanid */
void sql_do_chanmodes(char *chan, char **av)
{
    int atleastone = 0;
    char db[1000];              /* should be enough for long queries */
    char tmp[14] = "mode_XX=\"X\", ";
    char *modes;
    int argptr = 1;
    char buf[BUFSIZE];
    int nickid;
    char *user;
    int chanid;

    chan = rdb_escape(chan);
    chanid = db_getchannel(chan);

    if (!chanid) {
        denora_free(chan);
        return;
    }

    modes = av[0];

    SET_SEGV_LOCATION();
    snprintf(buf, sizeof(buf), "UPDATE %s SET ", ChanTable);

    if (*modes == '0') {
        return;
    }
    strcpy(db, buf);
    while (*modes) {
        switch (*modes) {
        case '+':
            tmp[9] = 'Y';
            break;
        case '-':
            tmp[9] = 'N';
            break;
        default:
            if (!strchr(ircd->cmodes, *modes)) {
                alog(LOG_DEBUG, langstr(ALOG_DEBUG_UNKNOW_CMODE), *modes);
                alog(LOG_DEBUG, langstr(ALOG_DEBUG_LAST_LINE), inbuf);
                alog(LOG_DEBUG, langstr(ALOG_DEBUG_REPORT_AT));
            } else if (*modes == ircd->ban_char) {
                argptr++;
            } else if (ircd->except && *modes == ircd->except_char) {
                argptr++;
            } else if (ircd->invitemode && *modes == ircd->invite_char) {
                argptr++;
            } else if (ircd->halfop && *modes == 'h') {
                SET_SEGV_LOCATION();
                user = rdb_escape(av[argptr++]);
                nickid = db_getnick(user);
                if (nickid) {
                    rdb_query
                        ("UPDATE %s SET mode_lh=\"%c\" WHERE chanid=%d AND nickid=%d",
                         IsOnTable, tmp[9], chanid, nickid);
                }
                denora_free(user);
            } else if (ircd->owner && *modes == 'q') {
                SET_SEGV_LOCATION();
                user = rdb_escape(av[argptr++]);
                nickid = db_getnick(user);
                if (nickid) {
                    rdb_query
                        ("UPDATE %s SET mode_lq=\"%c\" WHERE chanid=%d AND nickid=%d",
                         IsOnTable, tmp[9], chanid, nickid);
                }
                denora_free(user);
            } else if (ircd->protect && *modes == 'a') {
                SET_SEGV_LOCATION();
                user = rdb_escape(av[argptr++]);
                nickid = db_getnick(user);
                if (nickid) {
                    rdb_query
                        ("UPDATE %s SET mode_la=\"%c\" WHERE chanid=%d AND nickid=%d",
                         IsOnTable, tmp[9], chanid, nickid);
                }
                denora_free(user);
            } else if (*modes == 'o' || *modes == 'v') {
                SET_SEGV_LOCATION();
                user = rdb_escape(av[argptr++]);
                nickid = db_getnick(user);
                if (nickid) {
                    rdb_query
                        ("UPDATE %s SET mode_l%c=\"%c\" WHERE chanid=%d AND nickid=%d",
                         IsOnTable, *modes, tmp[9], chanid, nickid);
                }
                denora_free(user);
            } else {
                atleastone = 1;
                tmp[5] = ((*modes >= 'a') ? 'l' : 'u');
                tmp[6] = tolower(*modes);
                strcat(db, tmp);
                if (*modes == 'k') {
                    SET_SEGV_LOCATION();
                    if (tmp[9] == 'Y') {
                        char *key;
                        key = rdb_escape(av[argptr++]);
                        *buf = '\0';
                        snprintf(buf, BUFSIZE - 1, "mode_lk_data=\"%s\", ",
                                 (HidePasswords ? "HIDDEN" : key));
                        strcat(db, buf);
                        denora_free(key);
                    } else {
                        strcat(db, "mode_lk_data=\"\", ");
                        argptr++;       /* mode -k needs a parameter */
                    }
                } else if (*modes == 'l') {
                    SET_SEGV_LOCATION();
                    if (tmp[9] == 'Y') {
                        *buf = '\0';
                        snprintf(buf, BUFSIZE - 1, "mode_ll_data=\"%s\", ",
                                 av[argptr++]);
                        strcat(db, buf);
                    } else {
                        strcat(db, "mode_ll_data=\"\", ");
                    }
                } else if (ircd->Lmode && *modes == 'L') {
                    SET_SEGV_LOCATION();
                    if (tmp[9] == 'Y') {
                        char *ch = rdb_escape(av[argptr++]);
                        *buf = '\0';
                        snprintf(buf, BUFSIZE - 1, "mode_ul_data=\"%s\", ",
                                 ch);
                        strcat(db, buf);
                        denora_free(ch);
                    } else {
                        strcat(db, "mode_ul_data=\"\", ");
                    }
                } else if (ircd->fmode && ircd->floodchar
                           && *modes == ircd->floodchar) {
                    SET_SEGV_LOCATION();
                    if (tmp[9] == 'Y') {
                        *buf = '\0';
                        snprintf(buf, BUFSIZE - 1,
                                 "mode_l%c_data=\"%s\", ", ircd->floodchar,
                                 av[argptr++]);
                        strcat(db, buf);
                    } else {
                        *buf = '\0';
                        snprintf(buf, BUFSIZE - 1, "mode_l%c_data=\"\", ",
                                 ircd->floodchar);
                        strcat(db, buf);
                    }
                } else if (ircd->fmode && ircd->floodchar_alternative
                           && *modes == ircd->floodchar_alternative) {
                    SET_SEGV_LOCATION();
                    if (tmp[9] == 'Y') {
                        *buf = '\0';
                        snprintf(buf, BUFSIZE - 1,
                                 "mode_l%c_data=\"%s\", ",
                                 ircd->floodchar_alternative,
                                 av[argptr++]);
                        strcat(db, buf);
                    } else {
                        *buf = '\0';
                        snprintf(buf, BUFSIZE - 1, "mode_l%c_data=\"\", ",
                                 ircd->floodchar_alternative);
                        strcat(db, buf);
                    }
                }
            }
            break;
        }
        modes++;
    }
    if (atleastone) {
        sprintf(&db[strlen(db) - 2], " WHERE chanid=%d", chanid);
        rdb_query(db);
    }
    denora_free(chan);
}

/*************************************************************************/

/* PART */
void sql_do_part(char *chan, char *nick)
{
    int chanid, nickid, users;

    SET_SEGV_LOCATION();
    chan = rdb_escape(chan);
    nick = rdb_escape(nick);
    chanid = db_getchannel(chan);
    nickid = db_getnick(nick);

    if (chanid || nickid) {
        rdb_query("DELETE FROM %s WHERE nickid=%d AND chanid=%d",
                  IsOnTable, nickid, chanid);
        users = db_getchannel_users(chan);
        if (users) {
            users--;
            rdb_query("UPDATE %s SET currentusers=%d WHERE chanid=%d",
                      ChanTable, users, chanid);
        }
        db_checkemptychan(chanid);
    }
    SET_SEGV_LOCATION();
    denora_free(chan);
    denora_free(nick);
}

/*************************************************************************/

/* JOIN - join 0 */
void sql_do_partall(char *nick)
{
    nick = rdb_escape(nick);
    SET_SEGV_LOCATION();
    db_removefromchans(db_getnick(nick));
    denora_free(nick);
}

/*************************************************************************/

/* JOIN */
void sql_do_join(char *chan, char *nick)
{
    if (!denora->do_sql) {
        return;
    }
    chan = rdb_escape(chan);
    SET_SEGV_LOCATION();
    sql_do_addusers(db_getchancreate(chan), nick);
    denora_free(chan);
}

/*************************************************************************/

void sql_do_topic(char *chan, char *author, int topictime, char *topic)
{
    int chanid;

    if (!denora->do_sql) {
        return;
    }

    SET_SEGV_LOCATION();
    chan = rdb_escape(chan);
    author = rdb_escape(author);
    topic = (topic ? rdb_escape(topic) : sstrdup(""));
    chanid = db_getchannel(chan);
    SET_SEGV_LOCATION();
    if (chanid) {
        rdb_query
            ("UPDATE %s SET topic=\'%s\', topicauthor=\'%s\', topictime=FROM_UNIXTIME(%d) WHERE chanid=%d",
             ChanTable, topic, author, topictime, chanid);
    }
    denora_free(chan);
    denora_free(author);
    SET_SEGV_LOCATION();
    if (topic) {
        denora_free(topic);
    }
}

/*************************************************************************/

/* add one or more users to a chanid */
void sql_do_addusers(int chanid, char *users)
{
    int op, voice, halfop, owner, protect, shunned;
    char *nextusers;
    int nickid;
    User *u;

    SET_SEGV_LOCATION();

    while (users && (*users)) {
        /* Neostats is broken and sends an additional space. */
        if (*users == ' ')
            users++;

        /* Unreal uses SJOIN to send bans and exempts. Just ignore them. */
        if ((*users == ircd->sjoinbanchar)
            || (*users == ircd->sjoinexchar)) {
            nextusers = strchr(users, ' ');
            if (nextusers)
                *nextusers = '\0';
            users = nextusers;
            if (users)
                users++;
            continue;
        }
        op = 0;
        halfop = 0;
        voice = 0;
        owner = 0;
        protect = 0;
        shunned = 0;

        SET_SEGV_LOCATION();

        if (ircd->owner && !ircd->gagged) {
            if (*users == '!') {
                users++;
                owner = 1;
            }
            if ((*users == '*')) {
                users++;
                owner = 1;
            }
        }

        if (ircd->protect && !ircd->gagged && (*users == '!')) {
            users++;
            protect = 1;
        }

        if (ircd->gagged && (*users == '!')) {
            users++;
            shunned = 1;
        }

        if (ircd->halfop) {
            if (*users == '%') {
                users++;
                halfop = 1;
            }
        }
        SET_SEGV_LOCATION();

        if (*users == '@') {
            users++;
            op = 1;
            if (ircd->owner) {
                if (*users == '*') {
                    users++;
                    owner = 1;
                }
            }
        }
        if (ircd->halfop && (*users == '%')) {
            users++;
            halfop = 1;
        }

        if (*users == '+') {
            users++;
            voice = 1;
            if (ircd->owner) {
                if (*users == '*') {
                    users++;
                    owner = 1;
                }
            }
        }

        if (ircd->protect) {
            if (*users == '~') {
                users++;
                protect = 1;
            }
            if (ircd->owner) {
                if (*users == '*') {
                    users++;
                    owner = 1;
                }
            }
        }

        nextusers = strchr(users, ' ');
        if (nextusers)
            *nextusers = '\0';
        if (UseTS6) {
            u = find_byuid(users);
        } else {
            u = finduser(users);
        }
        users = rdb_escape((u ? u->nick : users));
        nickid = db_checknick((u ? u->nick : users));
        if (nickid != -1) {
            if (ircd->owner && ircd->protect && ircd->halfop
                && ircd->gagged) {
                rdb_query
                    ("INSERT IGNORE INTO %s (nickid, chanid, mode_lo, mode_lv, mode_lq, mode_lh, mode_la, mode_lg) VALUES (%d, %d,\'%s\',\'%s\',\'%s\',\'%s\',\'%s\', \'%s\')",
                     IsOnTable, nickid, chanid, (op ? "Y" : "N"),
                     (voice ? "Y" : "N"), (owner ? "Y" : "N"),
                     (halfop ? "Y" : "N"), (protect ? "Y" : "N"),
                     (shunned ? "Y" : "N"));
            }
            if (ircd->owner && ircd->protect && ircd->halfop
                && !ircd->gagged) {
                rdb_query
                    ("INSERT IGNORE INTO %s (nickid, chanid, mode_lo, mode_lv, mode_lq, mode_lh, mode_la) VALUES (%d, %d,\'%s\',\'%s\',\'%s\',\'%s\',\'%s\')",
                     IsOnTable, nickid, chanid, (op ? "Y" : "N"),
                     (voice ? "Y" : "N"), (owner ? "Y" : "N"),
                     (halfop ? "Y" : "N"), (protect ? "Y" : "N"));
            }

            if (!ircd->owner && ircd->protect && ircd->halfop
                && !ircd->gagged) {
                rdb_query
                    ("INSERT IGNORE INTO %s (nickid, chanid, mode_lo, mode_lv, mode_lh, mode_la) VALUES (%d, %d,\'%s\',\'%s\',\'%s\',\'%s\')",
                     IsOnTable, nickid, chanid, (op ? "Y" : "N"),
                     (voice ? "Y" : "N"), (halfop ? "Y" : "N"),
                     (protect ? "Y" : "N"));
            }

            if (!ircd->owner && !ircd->protect && ircd->halfop
                && !ircd->gagged) {
                rdb_query
                    ("INSERT IGNORE INTO %s (nickid, chanid, mode_lo, mode_lv, mode_lh) VALUES (%d, %d,\'%s\',\'%s\',\'%s\')",
                     IsOnTable, nickid, chanid, (op ? "Y" : "N"),
                     (voice ? "Y" : "N"), (halfop ? "Y" : "N"));
            }

            if (!ircd->owner && !ircd->protect && !ircd->halfop
                && !ircd->gagged) {
                rdb_query
                    ("INSERT IGNORE INTO %s (nickid, chanid, mode_lo, mode_lv) VALUES (%d, %d,\'%s\',\'%s\')",
                     IsOnTable, nickid, chanid, (op ? "Y" : "N"),
                     (voice ? "Y" : "N"));
            }
        } else {
            alog(LOG_NONEXISTANT,
                 langstr(ALOG_DEBUG_NONEXISTANT_USER_JOIN),
                 (u ? u->nick : users), chanid);
        }
        denora_free(users);
        users = nextusers;
        if (users) {
            users++;
        }
    }
}

/*************************************************************************/

/* SJOIN */
void sql_do_sjoin(char *chan, char *users, char **modes, int nbmodes)
{
    chan = rdb_escape(chan);
    SET_SEGV_LOCATION();
    sql_do_addusers(db_getchancreate(chan), users);
    if (nbmodes) {
        sql_do_chanmodes(chan, modes);
    }
    denora_free(chan);
}

/*************************************************************************/

void chan_deluser(User * user, Channel * c)
{
    struct c_userlist *u;

    SET_SEGV_LOCATION();

    for (u = c->users; u && u->user != user; u = u->next);
    if (!u)
        return;

    if (u->ud) {
        denora_free(u->ud);
    }

    if (u->next) {
        u->next->prev = u->prev;
    }
    if (u->prev) {
        u->prev->next = u->next;
    } else {
        c->users = u->next;
    }
    SET_SEGV_LOCATION();
    denora_free(u);
    c->stats->usercount--;
    SET_SEGV_LOCATION();

    if (!c->users)
        chan_delete(c);
}

/*************************************************************************/

/* Returns a fully featured binary modes string. If complete is 0, the
 * eventual parameters won't be added to the string.
 */
char *chan_get_modes(Channel * chan, int complete, int plus)
{
    static char res[BUFSIZE];
    char *end = res;

    SET_SEGV_LOCATION();

    if (chan->mode) {
        int n = 0;
        CBModeInfo *cbmi = cbmodeinfos;

        do {
            if (chan->mode & cbmi->flag)
                *end++ = cbmi->mode;
        } while ((++cbmi)->flag != 0 && ++n < (int) sizeof(res) - 1);

        if (complete) {
            cbmi = cbmodeinfos;
            SET_SEGV_LOCATION();

            do {
                if (cbmi->getvalue && (chan->mode & cbmi->flag) && plus) {
                    char *value = cbmi->getvalue(chan);

                    if (value) {
                        *end++ = ' ';
                        while (*value)
                            *end++ = *value++;
                    }
                }
            } while ((++cbmi)->flag != 0 && ++n < (int) sizeof(res) - 1);
        }
    }

    *end = 0;

    return res;
}

/*************************************************************************/

/* Remove the status of an user on a channel */
void chan_remove_user_status(Channel * chan, User * user, int16 status)
{
    struct u_chanlist *uc;

    SET_SEGV_LOCATION();

    for (uc = user->chans; uc; uc = uc->next) {
        if (uc->chan == chan) {
            uc->status &= ~status;
            break;
        }
    }
}

/*************************************************************************/

void chan_set_modes(Channel * chan, int ac, char **av)
{
    int add = 1;
    char *modes = av[0], mode;
    CBMode *cbm;
    CMMode *cmm;

    alog(LOG_DEBUG, langstr(ALOG_DEBUG_CHANGE_MODE_CHAN), chan->name,
         merge_args(ac, av));

    ac--;
    SET_SEGV_LOCATION();

    while ((mode = *modes++)) {

        switch (mode) {
        case '+':
            add = 1;
            continue;
        case '-':
            add = 0;
            continue;
        }

        if (((int) mode) < 0) {
            alog(LOG_DEBUG, langstr(AL0G_DEBUG_BAD_CHAN_MODE), chan->name);
            continue;
        }
        SET_SEGV_LOCATION();

        if ((cbm = &cbmodes[(int) mode])->flag != 0) {
            if (add)
                chan->mode |= cbm->flag;
            else
                chan->mode &= ~cbm->flag;

            if (cbm->setvalue) {
                cbm->setvalue(chan, add ? *av : NULL);
            }
        } else if ((cmm = &cmmodes[(int) mode])->addmask) {
            if (ac == 0) {
                alog(LOG_ERROR, langstr(ALOG_DEBUG_MODE_NO_PARAM),
                     add ? '+' : '-', mode, chan->name);
                continue;
            }

            ac--;
            av++;
            add ? cmm->addmask(chan, *av) : cmm->delmask(chan, *av);
        }
    }

}

/*************************************************************************/

/* Return the Channel structure corresponding to the named channel, or NULL
 * if the channel was not found.  chan is assumed to be non-NULL and valid
 * (i.e. pointing to a channel name of 2 or more characters). 
 */
Channel *findchan(const char *chan)
{
    Channel *c;

    SET_SEGV_LOCATION();

    if (!chan || !*chan) {
        alog(LOG_ERROR, langstr(ALOG_DEBUG_FINDCHAN_BAD));
        do_backtrace(0);
        return NULL;
    }

    alog(LOG_EXTRADEBUG, "debug: findchan(%p)", chan);
    c = chanlist[HASH(chan)];
    SET_SEGV_LOCATION();
    while (c) {
        if (stricmp(c->name, chan) == 0) {
            return c;
        }
        c = c->next;
    }
    alog(LOG_EXTRADEBUG, "debug: findchan(%s) -> %p", chan, (void *) c);
    return NULL;
}

/*************************************************************************/

StatsChannel *findstatschan(const char *chan)
{
    StatsChannel *c;

    SET_SEGV_LOCATION();

    if (!chan || !*chan) {
        alog(LOG_ERROR, langstr(ALOG_DEBUG_FINDSTATCHAN_BAD));
        return NULL;
    }

    alog(LOG_EXTRADEBUG, "debug: findstatschan(%p)", chan);
    c = statschanlist[STATSHASH(chan)];
    SET_SEGV_LOCATION();
    while (c) {
        if (stricmp(c->name, chan) == 0) {
            return c;
        }
        c = c->next;
    }
    alog(LOG_EXTRADEBUG, "debug: findstatschan(%s) -> %p", chan,
         (void *) c);
    return NULL;
}

/*************************************************************************/

/* Iterate over all channels in the channel list.  Return NULL at end of
 * list.
 */
Channel *firstchan(void)
{
    next_index = 0;
    SET_SEGV_LOCATION();

    while (next_index < 1024 && current == NULL) {
        current = chanlist[next_index++];
    }
    SET_SEGV_LOCATION();
    alog(LOG_EXTRADEBUG, "debug: firstchan() returning %s",
         current ? current->name : "NULL (end of list)");
    return current;
}

/*************************************************************************/

StatsChannel *first_statschan(void)
{
    next_index = 0;
    SET_SEGV_LOCATION();

    while (next_index < 1024 && currentsc == NULL) {
        currentsc = statschanlist[next_index++];
    }
    alog(LOG_EXTRADEBUG, "debug: first_statschan() returning %s",
         currentsc ? currentsc->name : "NULL (end of list)");
    SET_SEGV_LOCATION();
    return currentsc;
}

/*************************************************************************/

Channel *nextchan(void)
{
    SET_SEGV_LOCATION();

    if (current) {
        current = current->next;
    }
    if (!current && next_index < 1024) {
        while (next_index < 1024 && current == NULL)
            current = chanlist[next_index++];
    }
    SET_SEGV_LOCATION();

    alog(LOG_EXTRADEBUG, "debug: nextchan() returning %s",
         current ? current->name : "NULL (end of list)");
    return current;
}

/*************************************************************************/

StatsChannel *next_statschan(void)
{
    SET_SEGV_LOCATION();

    if (currentsc)
        currentsc = currentsc->next;
    if (!currentsc && next_index < 1024) {
        while (next_index < 1024 && currentsc == NULL) {
            currentsc = statschanlist[next_index++];
        }
    }
    alog(LOG_EXTRADEBUG, "debug: next_statschan() returning %s",
         currentsc ? currentsc->name : "NULL (end of list)");
    return currentsc;
}

/*************************************************************************/

void do_bmask(char **av)
{
    Channel *c;
    char *bans;
    char *b;
    int count, i;

    c = findchan(av[1]);

    if (c) {
        bans = sstrdup(av[3]);
        count = myNumToken(bans, ' ');
        for (i = 0; i <= count - 1; i++) {
            b = myStrGetToken(bans, ' ', i);
            if (!stricmp(av[2], "b")) {
                add_ban(c, b);
            }
            if (!stricmp(av[2], "e")) {
                add_exception(c, b);
            }
            if (!stricmp(av[2], "I")) {
                add_invite(c, b);
            }
            denora_free(b);
        }
        denora_free(bans);
    }
}

/*************************************************************************/

void do_p10_burst(int ac, char **av)
{
    Channel *c;
    User *user;
    char *s, *t, *m;
    char *v[32];
    char *x[32];
    int i = 0;

    /* Double check to avoid unknown modes that need parameters */
    if (ac == 5) {
        c = findchan(av[0]);
        while ((s = myStrGetToken(av[3], ',', i))) {
            t = myStrGetToken(s, ':', 0);
            m = myStrGetToken(s, ':', 1);
            user = find_byuid(t);
            if (!user) {
                alog(LOG_NONEXISTANT,
                     langstr(ALOG_DEBUG_SJOIN_NONEXISTANT), t, av[0]);
                i++;
                continue;
            }
            v[0] = av[0];
            do_join(user->nick, 1, v);
            if (m) {
                x[0] = av[0];
                if (!stricmp(m, "o")) {
                    x[1] = sstrdup("+o");
                } else if (!stricmp(m, "v")) {
                    x[1] = sstrdup("+v");
                } else if (!stricmp(m, "h") && ircd->halfop) {
                    x[1] = sstrdup("+h");
                } else if (!stricmp(m, "ov")) {
                    x[1] = sstrdup("+ov");
                } else if (!stricmp(m, "hv")) {
                    x[1] = sstrdup("+hv");
                } else if (!stricmp(m, "oh")) {
                    x[1] = sstrdup("+oh");
                } else if (!stricmp(m, "ohv")) {
                    x[1] = sstrdup("+ohv");
                }
                x[2] = user->nick;
                do_cmode(user->nick, 3, x);
                denora_free(m);
                denora_free(x[1]);
            }
            i++;
            denora_free(t);
        }
        if (c) {
            /* Set the timestamp */
            c->creation_time = strtol(av[1], NULL, 10);
            /* We now update the channel mode. */
            chan_set_modes(c, 1, &av[2]);
        }
        /* all that and now to the bans */
        c = findchan(av[0]);
        i = 0;
        if (c) {
            while ((s = myStrGetToken(av[4], ' ', i))) {
                add_ban(c, s);
                i++;
            }
        }
    } else if (ac == 3) {
        c = findchan(av[0]);
        while ((s = myStrGetToken(av[2], ',', i))) {
            t = myStrGetToken(s, ':', 0);
            m = myStrGetToken(s, ':', 1);
            user = find_byuid(t);
            if (!user) {
                alog(LOG_NONEXISTANT,
                     langstr(ALOG_DEBUG_SJOIN_NONEXISTANT), t, av[0]);
                i++;
                continue;
            }
            v[0] = av[0];
            do_join(user->nick, 1, v);
            if (m) {
                x[0] = av[0];
                if (!stricmp(m, "o")) {
                    x[1] = sstrdup("+o");
                } else if (!stricmp(m, "v")) {
                    x[1] = sstrdup("+v");
                } else if (!stricmp(m, "h") && ircd->halfop) {
                    x[1] = sstrdup("+h");
                } else if (!stricmp(m, "ov")) {
                    x[1] = sstrdup("+ov");
                } else if (!stricmp(m, "hv")) {
                    x[1] = sstrdup("+hv");
                } else if (!stricmp(m, "oh")) {
                    x[1] = sstrdup("+oh");
                } else if (!stricmp(m, "ohv")) {
                    x[1] = sstrdup("+ohv");
                }
                x[2] = user->nick;
                do_cmode(user->nick, 3, x);
                denora_free(m);
                denora_free(x[1]);
            }
            i++;
        }
    } else if (ac == 4) {
        c = findchan(av[0]);
        if (*av[2] == '+') {
            while ((s = myStrGetToken(av[3], ',', i))) {
                t = myStrGetToken(s, ':', 0);
                m = myStrGetToken(s, ':', 1);
                user = find_byuid(t);
                if (!user) {
                    alog(LOG_NONEXISTANT,
                         langstr(ALOG_DEBUG_SJOIN_NONEXISTANT), t, av[0]);
                    i++;
                    continue;
                }
                v[0] = av[0];
                do_join(user->nick, 1, v);
                if (m) {
                    x[0] = av[0];
                    if (!stricmp(m, "o")) {
                        x[1] = sstrdup("+o");
                    } else if (!stricmp(m, "v")) {
                        x[1] = sstrdup("+v");
                    } else if (!stricmp(m, "h") && ircd->halfop) {
                        x[1] = sstrdup("+h");
                    } else if (!stricmp(m, "ov")) {
                        x[1] = sstrdup("+ov");
                    } else if (!stricmp(m, "hv")) {
                        x[1] = sstrdup("+hv");
                    } else if (!stricmp(m, "oh")) {
                        x[1] = sstrdup("+oh");
                    } else if (!stricmp(m, "ohv")) {
                        x[1] = sstrdup("+ohv");
                    }
                    x[2] = user->nick;
                    do_cmode(user->nick, 3, x);
                    denora_free(m);
                    denora_free(x[1]);
                }
                i++;
                if (c) {
                    /* Set the timestamp */
                    c->creation_time = strtol(av[1], NULL, 10);
                    /* We now update the channel mode. */
                    chan_set_modes(c, 1, &av[2]);
                }
            }
        } else {
            while ((s = myStrGetToken(av[2], ',', i))) {
                t = myStrGetToken(s, ':', 0);
                m = myStrGetToken(s, ':', 1);
                user = find_byuid(t);
                if (!user) {
                    alog(LOG_NONEXISTANT,
                         langstr(ALOG_DEBUG_SJOIN_NONEXISTANT), t, av[0]);
                    i++;
                    continue;
                }
                v[0] = av[0];
                do_join(user->nick, 1, v);
                if (m) {
                    x[0] = av[0];
                    if (!stricmp(m, "o")) {
                        x[1] = sstrdup("+o");
                    } else if (!stricmp(m, "v")) {
                        x[1] = sstrdup("+v");
                    } else if (!stricmp(m, "h") && ircd->halfop) {
                        x[1] = sstrdup("+h");
                    } else if (!stricmp(m, "ov")) {
                        x[1] = sstrdup("+ov");
                    } else if (!stricmp(m, "hv")) {
                        x[1] = sstrdup("+hv");
                    } else if (!stricmp(m, "oh")) {
                        x[1] = sstrdup("+oh");
                    } else if (!stricmp(m, "ohv")) {
                        x[1] = sstrdup("+ohv");
                    }
                    x[2] = user->nick;
                    do_cmode(user->nick, 3, x);
                    denora_free(m);
                    denora_free(x[1]);
                }
                i++;
            }
            /* all that and now to the bans */
            c = findchan(av[0]);
            i = 0;
            if (c) {
                while ((s = myStrGetToken(av[3], ' ', i))) {
                    add_ban(c, s);
                    i++;
                }
            }
        }
    }
}


/* Handle a JOIN command.
 *	av[0] = channels to join
 */
void do_join(const char *source, int ac, char **av)
{
    User *user;
    char *s, *t;
    struct u_chanlist *c, *nextc;

    SET_SEGV_LOCATION();

    if (ircd->p10) {
        user = find_byuid(source);
        if (!user) {
            user = finduser(source);
        }
    } else {
        user = finduser(source);
    }
    if (!user) {
        alog(LOG_NONEXISTANT, langstr(ALOG_DEBUG_JOIN_NONEXISTANT),
             source, merge_args(ac, av));
        return;
    }

    t = av[0];
    while (*(s = t)) {
        t = s + strcspn(s, ",");
        if (*t)
            *t++ = 0;
        alog(LOG_DEBUG, langstr(ALOG_DEBUG_JOIN), source, s);
        SET_SEGV_LOCATION();

        if (*s == '0') {
            c = user->chans;
            while (c) {
                nextc = c->next;
                chan_deluser(user, c->chan);
                if (denora->do_sql)
                    sql_do_partall(user->nick);
                send_event(EVENT_USER_PART, 3, user->nick, c->chan->name,
                           "Left all channels");
                denora_free(c);
                c = nextc;
            }
            user->chans = NULL;
            continue;
        }
        join_user_update(user, findchan(s), s);
        if (denora->do_sql) {
            sql_do_join(s, user->nick);
        }
        SET_SEGV_LOCATION();
    }
}

/*************************************************************************/

void do_p10_kick(const char *source, int ac, char **av)
{
    char *v[32];
    User *u, *u2;

    if (ac != 3) {
        return;
    }

    u = find_byuid(source);
    u2 = find_byuid(av[1]);

    v[0] = sstrdup(av[0]);
    v[1] = sstrdup((u2 ? u2->nick : av[1]));
    v[2] = sstrdup(av[1]);
    do_kick((u ? u->nick : source), ac, v);
    denora_free(v[0]);
    denora_free(v[1]);
    denora_free(v[2]);
}

/* Handle a KICK command.
 *	av[0] = channel
 *	av[1] = nick(s) being kicked
 *	av[2] = reason
 */
void do_kick(const char *source, int ac, char **av)
{
    User *user, *kicker;
    char *s, *t, *chan;
    struct u_chanlist *c;
    Channel *c2;
    ChannelStats *cs;

    SET_SEGV_LOCATION();

    t = av[1];
    while (*(s = t)) {
        t = s + strcspn(s, ",");
        if (*t)
            *t++ = 0;

        user = finduser(s);
        if (!user) {
            alog(LOG_NONEXISTANT, langstr(ALOG_DEBUG_KICK_NONEXISTANT), s,
                 av[0], merge_args(ac - 2, av + 2));
            continue;
        }
        alog(LOG_DEBUG, langstr(ALOG_DEBUG_KICKED), s, av[0]);

        c2 = findchan(av[0]);
        if (c2) {
            c2->stats->kickcount++;
            chan = rdb_escape(c2->name);
            rdb_query("UPDATE %s SET kickcount=%d WHERE chanid=%d",
                      ChanTable, c2->stats->kickcount,
                      db_getchannel(chan));
            denora_free(chan);
        }
        if ((cs = find_cs(av[0]))) {
            if (!stricmp(s, s_StatServ)) {
                denora_cmd_join(s_StatServ, av[0], time(NULL));
                return;
            }
        }

        kicker = finduser(source);
        /* if we can't find the person whom kicked they are likely
           a server or nonexistent either way we shouldn't count them */
        if (kicker && denora->do_sql && !LargeNet) {
            count_kicks(kicker, user, c2);
        }
        SET_SEGV_LOCATION();

        for (c = user->chans; c && stricmp(av[0], c->chan->name) != 0;
             c = c->next);
        if (c) {
            chan_deluser(user, c->chan);
            SET_SEGV_LOCATION();
            if (c->next) {
                c->next->prev = c->prev;
            }
            if (c->prev) {
                c->prev->next = c->next;
            } else {
                user->chans = c->next;
            }
            SET_SEGV_LOCATION();
            denora_free(c);
        }
    }
}

/*************************************************************************/

/* Handle a PART command.
 *	av[0] = channels to leave
 *	av[1] = reason (optional)
 */
void do_part(const char *source, int ac, char **av)
{
    User *user;
    char *s, *t;
    struct u_chanlist *c;

    SET_SEGV_LOCATION();

    user = finduser(source);
    if (!user) {
        alog(LOG_NONEXISTANT, langstr(ALOG_DEBUG_PART_NONEXISTANT),
             source, merge_args(ac, av));
        return;
    }
    SET_SEGV_LOCATION();
    t = av[0];
    while (*(s = t)) {
        t = s + strcspn(s, ",");
        if (*t)
            *t++ = 0;
        alog(LOG_DEBUG, langstr(ALOG_DEBUG_PART), source, s);

        for (c = user->chans; c && stricmp(s, c->chan->name) != 0;
             c = c->next);
        if (c) {
            if (!c->chan) {
                alog(LOG_ERROR, langstr(ALOG_DEBUG_PART_BUG), s);
                return;
            }
            SET_SEGV_LOCATION();
            chan_deluser(user, c->chan);
            if (denora->do_sql)
                sql_do_part(s, user->nick);
            if (c->next) {
                c->next->prev = c->prev;
            }
            if (c->prev) {
                c->prev->next = c->next;
            } else {
                user->chans = c->next;
            }
            SET_SEGV_LOCATION();
            send_event(EVENT_USER_PART, 3, user->nick, c->chan->name,
                       av[1] ? av[1] : "");
            denora_free(c);
        }
    }
}

/*************************************************************************/

/* Handle a SJOIN command.

   On channel creation, syntax is:

   av[0] = timestamp
   av[1] = channel name
   av[2|3|4] = modes   \   depends of whether the modes k and l
   av[3|4|5] = users   /   are set or not.

   When a single user joins an (existing) channel, it is:

   av[0] = timestamp
   av[1] = user

   ============================================================

   Unreal SJOIN

   On Services connect there is
   SJOIN !11LkOb #ircops +nt :@Trystan &*!*@*.aol.com "*@*.home.com

   av[0] = time stamp (base64)
   av[1] = channel
   av[2] = modes
   av[3] = users + bans + exceptions

   On Channel Creation or a User joins an existing
   Luna.NomadIrc.Net SJOIN !11LkW9 #akill :@Trystan
   Luna.NomadIrc.Net SJOIN !11LkW9 #akill :Trystan`

   av[0] = time stamp (base64)
   av[1] = channel
   av[2] = users

*/
void do_sjoin(const char *source, int ac, char **av)
{
    Channel *c;
    User *user;
    char *s, *end, cubuf[7], *end2, *sumodes[6];
    char *sqlusers = NULL;
    char *mask;

    int ts = 0;

    SET_SEGV_LOCATION();

    if (ircd->sjb64) {
        ts = base64dec_ts(av[0]);
    } else {
        ts = strtoul(av[0], NULL, 10);
    }

    /* Double check to avoid unknown modes that need parameters */
    if (ac >= 4 && ac <= 6) {
        if (denora->do_sql) {
            sqlusers = sstrdup(av[ac - 1]);
            if (denora->do_sql)
                sql_do_sjoin(av[1], sqlusers, &av[2],
                             (ac > 3) ? ac - 3 : 0);
            denora_free(sqlusers);
        }
        c = findchan(av[1]);

        cubuf[0] = '+';
        sumodes[0] = cubuf;

        /* We make all the users join */
        s = av[ac - 1];         /* Users are always the last element */

        while (*s) {
            end = strchr(s, ' ');
            if (end)
                *end = 0;

            end2 = cubuf + 1;


            if (ircd->sjoinbanchar) {
                if (*s == ircd->sjoinbanchar) {
                    mask = myStrGetToken(s, ircd->sjoinbanchar, 1);
                    add_ban(c, mask);
                    denora_free(mask);
                    if (!end)
                        break;
                    s = end + 1;
                    continue;
                }
            }
            if (ircd->sjoinexchar) {
                if (*s == ircd->sjoinexchar) {
                    mask = myStrGetToken(s, ircd->sjoinexchar, 1);
                    add_exception(c, mask);
                    denora_free(mask);
                    if (!end)
                        break;
                    s = end + 1;
                    continue;
                }
            }

            if (*s == '\'') {
                mask = myStrGetToken(s, '\'', 1);
                add_invite(c, mask);
                denora_free(mask);
                if (!end)
                    break;
                s = end + 1;
                continue;
            }
            while (csmodes[(int) *s] != 0)
                *end2++ = csmodes[(int) *s++];
            *end2 = 0;

            if (UseTS6) {
                user = find_byuid(s);
                if (!user) {
                    user = finduser(s);
                }
            } else {
                user = finduser(s);
            }
            if (!user) {
                alog(LOG_NONEXISTANT,
                     langstr(ALOG_DEBUG_SJOIN_NONEXISTANT), s, av[1]);
                return;
            }

            c = join_user_update(user, c, av[1]);
            /* We update user mode on the channel */
            if (end2 - cubuf > 1) {
                int i;

                for (i = 1; i < end2 - cubuf; i++)
                    sumodes[i] = user->nick;
                chan_set_modes(c, 1 + (end2 - cubuf - 1), sumodes);
            }

            if (!end)
                break;
            s = end + 1;
        }

        if (c) {
            /* Set the timestamp */
            c->creation_time = ts;
            /* We now update the channel mode. */
            chan_set_modes(c, ac - 3, &av[2]);
            SET_SEGV_LOCATION();
        }

        /* Unreal just had to be different */
    } else if (ac == 3) {
        if (denora->do_sql) {
            sqlusers = sstrdup(av[ac - 1]);
            if (denora->do_sql)
                sql_do_sjoin(av[1], sqlusers, &av[2],
                             (ac > 3) ? ac - 3 : 0);
            denora_free(sqlusers);
        }
        c = findchan(av[1]);
        cubuf[0] = '+';
        sumodes[0] = cubuf;

        /* We make all the users join */
        s = av[2];              /* Users are always the last element */

        while (*s) {
            end = strchr(s, ' ');
            if (end)
                *end = 0;

            end2 = cubuf + 1;

            while (csmodes[(int) *s] != 0)
                *end2++ = csmodes[(int) *s++];
            *end2 = 0;

            if (UseTS6) {
                user = find_byuid(s);
                if (!user) {
                    user = finduser(s);
                }
            } else {
                user = finduser(s);
            }
            if (!user) {
                alog(LOG_NONEXISTANT,
                     langstr(ALOG_DEBUG_SJOIN_NONEXISTANT), s, av[1]);
                return;
            }

            c = join_user_update(user, c, av[1]);
            SET_SEGV_LOCATION();

            /* We update user mode on the channel */
            if (end2 - cubuf > 1) {
                int i;

                for (i = 1; i < end2 - cubuf; i++) {
                    sumodes[i] = user->nick;
                }
                chan_set_modes(c, 1 + (end2 - cubuf - 1), sumodes);
                SET_SEGV_LOCATION();
            }

            if (!end)
                break;
            s = end + 1;
        }

        if (c) {
            c->creation_time = ts;
        }
    } else if (ac == 2) {
        if (denora->do_sql) {
            sqlusers = sstrdup(source);
            if (denora->do_sql)
                sql_do_join(av[1], sqlusers);
            denora_free(sqlusers);
        }
        if (UseTS6) {
            user = find_byuid(source);
            if (!user) {
                user = finduser(source);
            }
        } else {
            user = finduser(source);
        }
        if (!user) {
            alog(LOG_NONEXISTANT, langstr(ALOG_DEBUG_SJOIN_NONEXISTANT),
                 source, av[1]);
            return;
        }
        SET_SEGV_LOCATION();
        c = findchan(av[1]);
        c = join_user_update(user, c, av[1]);
        c->creation_time = ts;
    }
}


/*************************************************************************/

/* Handle a channel MODE command. */
void do_cmode(const char *source, int ac, char **av)
{
    Channel *chan;
    int i;
    char *t;
    User *u;
    Exclude *e;

    SET_SEGV_LOCATION();

    if (ircdcap->tsmode) {
        /* TSMODE for bahamut - leave this code out to break MODEs. -GD */
        /* if they don't send it in CAPAB check if we just want to enable it */
        if (denora->capab & ircdcap->tsmode || UseTSMODE) {
            for (i = 0; i < (int) strlen(av[1]); i++) {
                if (!isdigit(av[1][i]))
                    break;
            }
            if (av[1][i] == '\0') {
                /* We have a valid TS field in av[1] now, so we can strip it off */
                /* After we swap av[0] and av[1] of course to not break stuff! :) */
                t = av[0];
                av[0] = av[1];
                av[1] = t;
                ac--;
                av++;
            } else {
                alog(LOG_ERROR, langstr(ALOG_NON_TS_MODE));
            }
        }
    }

    chan = findchan(av[0]);
    if (!chan) {
        alog(LOG_DEBUG, langstr(ALOG_DEBUG_NO_FIND_CHAN), av[0]);
        return;
    }
    SET_SEGV_LOCATION();

    /* This shouldn't trigger on +o, etc. */
    if (strchr(source, '.') && !av[1][strcspn(av[1], "bovahq")]) {
        if (time(NULL) != chan->server_modetime) {
            chan->server_modecount = 0;
            chan->server_modetime = time(NULL);
        }
        chan->server_modecount++;
    }

    ac--;
    av++;
    SET_SEGV_LOCATION();
    chan_set_modes(chan, ac, av);
    SET_SEGV_LOCATION();
    if (denora->do_sql)
        sql_do_chanmodes(chan->name, av);
    if ((u = finduser(source))) {
        if (denora->do_sql && !LargeNet) {
            e = find_exclude(u->nick, u->server->name);
            if (!e) {
                count_modes(u, chan);
            }
        }
    }
}

/*************************************************************************/

/* Handle a TOPIC command. */

void do_topic(int ac, char **av)
{
    Channel *c = findchan(av[0]);
    int ts;
    time_t topic_time;
    User *u;
    Exclude *e;
    char *s, *setter;

    SET_SEGV_LOCATION();

    if (ircd->sjb64) {
        ts = base64dec_ts(av[2]);
        alog(LOG_DEBUG, langstr(ALOG_DEBUG_TOPIC_TS), av[2], ts);
    } else {
        if (ac == 2) {
            ts = time(NULL);
        } else {
            ts = strtoul(av[2], NULL, 10);
        }
    }

    topic_time = ts;

    if (!c) {
        alog(LOG_NONEXISTANT, langstr(ALOG_TOPIC_NONEXISTANT),
             merge_args(ac - 1, av + 1), av[0]);
        return;
    }
    SET_SEGV_LOCATION();

    if (c->topic) {
        denora_free(c->topic);
        c->topic = NULL;
    }
    if (ac > 3 && *av[3]) {
        c->topic = sstrdup(av[3]);
    }
    if (av[1]) {
        setter = sstrdup(av[1]);
    } else {
        setter = sstrdup("");
    }
    s = myStrGetToken(setter, '!', 0);
    if (s) {
        strscpy(c->topic_setter, s, sizeof(c->topic_setter));
        denora_free(s);
    } else {
        strscpy(c->topic_setter, setter, sizeof(c->topic_setter));
    }
    denora_free(setter);
    c->topic_time = topic_time;
    c->stats->topic_count++;
    sql_do_topic(c->name, c->topic_setter, topic_time, c->topic);
    SET_SEGV_LOCATION();

    if (av[1]) {
        if ((u = finduser(c->topic_setter))) {
            if (denora->do_sql && !LargeNet) {
                e = find_exclude(u->nick, u->server->name);
                if (!e) {
                    count_topics(u, c);
                }
            }
        }
    }
}

/*************************************************************************/

void add_ban(Channel * chan, char *mask)
{
    SET_SEGV_LOCATION();

    /* check for NULL values otherwise we will segfault */
    if (!chan || !mask) {
        alog(LOG_DEBUG, langstr(ALOG_ADD_BAN_BAD));
        return;
    }

    if (chan->bancount >= chan->bansize) {
        chan->bansize += 8;
        chan->bans = srealloc(chan->bans, sizeof(char *) * chan->bansize);
    }
    chan->bans[chan->bancount++] = sstrdup(mask);

    SET_SEGV_LOCATION();

    alog(LOG_DEBUG, langstr(ALOG_ADD_BAN_ADDED), mask, chan->name);

    if (denora->do_sql) {
        sql_channel_ban(ADD, chan->name, mask);
    }
}

/*************************************************************************/

void add_exception(Channel * chan, char *mask)
{
    SET_SEGV_LOCATION();

    if (chan->exceptcount >= chan->exceptsize) {
        chan->exceptsize += 8;
        chan->excepts =
            srealloc(chan->excepts, sizeof(char *) * chan->exceptsize);
    }
    chan->excepts[chan->exceptcount++] = sstrdup(mask);

    SET_SEGV_LOCATION();

    alog(LOG_DEBUG, langstr(ALOG_ADD_EXCEPTION_ADDED), mask, chan->name);

    if (denora->do_sql) {
        sql_channel_exception(ADD, chan->name, mask);
    }
}

/*************************************************************************/

void add_invite(Channel * chan, char *mask)
{
    SET_SEGV_LOCATION();

    if (chan->invitecount >= chan->invitesize) {
        chan->invitesize += 8;
        chan->invite =
            srealloc(chan->invite, sizeof(char *) * chan->invitesize);
    }
    chan->invite[chan->invitecount++] = sstrdup(mask);

    alog(LOG_DEBUG, langstr(ALOG_ADD_INVITE_ADDED), mask, chan->name);
    SET_SEGV_LOCATION();

    if (denora->do_sql) {
        sql_channel_invite(ADD, chan->name, mask);
    }
}

/*************************************************************************/

/* Add/remove a user to/from a channel, creating or deleting the channel as
 * necessary.  If creating the channel, restore mode lock and topic as
 * necessary.  Also check for auto-opping and auto-voicing.
 * Modified, so ignored users won't get any status via services -certus 
 */
void chan_adduser2(User * user, Channel * c)
{
    struct c_userlist *u;
    char *chan;

    SET_SEGV_LOCATION();

    u = scalloc(sizeof(struct c_userlist), 1);
    u->next = c->users;
    if (c->users) {
        c->users->prev = u;
    }
    c->users = u;
    u->user = user;
    c->stats->usercount++;
    if (!c->stats->maxusertime) {
        c->stats->maxusertime = time(NULL);
    }
    if (c->stats->usercount > c->stats->maxusercount) {
        c->stats->maxusercount++;
        c->stats->maxusertime = time(NULL);
    }
    if (denora->do_sql) {
        chan = rdb_escape(c->name);
        rdb_query
            ("UPDATE %s SET currentusers=%d, maxusers=%d, maxusertime=%ld WHERE chanid=%d",
             ChanTable, c->stats->usercount, c->stats->maxusercount,
             (long int) c->stats->maxusertime, db_getchancreate(chan));
        denora_free(chan);
    }
    c->stats->joincounter++;
    send_event(EVENT_USER_JOIN, 2, user->nick, c->name);
    SET_SEGV_LOCATION();
}

/*************************************************************************/

/* This creates the channel structure (was originally in
 *  chan_adduser, but splitted to make it more efficient to use for
 *  SJOINs). 
 */
Channel *chan_create(char *chan)
{
    Channel *c;
    StatsChannel *sc;
    Channel **list;

    SET_SEGV_LOCATION();

    alog(LOG_DEBUG, langstr(ALOG_CREATE_CHAN), chan);

    /* Allocate pre-cleared memory */
    c = scalloc(sizeof(Channel), 1);
    strscpy(c->name, chan, sizeof(c->name));
    list = &chanlist[HASH(c->name)];
    c->next = *list;
    if (*list) {
        (*list)->prev = c;
    }
    SET_SEGV_LOCATION();

    *list = c;
    c->creation_time = time(NULL);
    c->cstats = 0;
    sc = findstatschan(c->name);
    if (sc) {
        c->stats = sc;
    } else {
        sc = statschan_create(chan);
        c->stats = sc;
    }
    SET_SEGV_LOCATION();

    c->stats->in_use = 1;
    stats->chans++;
    do_checkchansmax();
    chans_hourly(NULL);
    return c;
}

/*************************************************************************/

StatsChannel *statschan_create(char *chan)
{
    StatsChannel *c, **list;

    SET_SEGV_LOCATION();

    /* Allocate pre-cleared memory */
    c = scalloc(sizeof(StatsChannel), 1);
    strscpy(c->name, chan, sizeof(c->name));
    list = &statschanlist[STATSHASH(c->name)];
    c->next = *list;
    if (*list) {
        (*list)->prev = c;
    }
    *list = c;
    SET_SEGV_LOCATION();

    return c;
}

/*************************************************************************/

/* This destroys the channel structure, freeing everything in it. */

void chan_delete(Channel * c)
{
    BanData *bd, *next;
    int i;

    SET_SEGV_LOCATION();

    alog(LOG_EXTRADEBUG, langstr(ALOG_DELETE_CHAN), c->name);

    c->stats->in_use = 0;
    stats->chans--;

    for (bd = c->bd; bd; bd = next) {
        if (bd->mask) {
            denora_free(bd->mask);
        }
        next = bd->next;
        denora_free(bd);
    }

    if (c->topic) {
        denora_free(c->topic);
    }
    SET_SEGV_LOCATION();

    if (c->key) {
        denora_free(c->key);
    }
    if (ircd->fmode) {
        if (c->flood) {
            denora_free(c->flood);
        }
    }
    if (ircd->Lmode) {
        if (c->redirect) {
            denora_free(c->redirect);
        }
    }
    SET_SEGV_LOCATION();

    for (i = 0; i < c->bancount; ++i) {
        if (c->bans[i]) {
            denora_free(c->bans[i]);
        } else {
            alog(LOG_ERROR, langstr(ALOG_BAN_FREE_ERROR), c->name, i);
        }
    }
    if (c->bansize) {
        denora_free(c->bans);
    }

    if (denora->do_sql)
        sql_channel_ban(ALL, c->name, NULL);
    SET_SEGV_LOCATION();

    if (ircd->except) {
        for (i = 0; i < c->exceptcount; ++i) {
            if (c->excepts[i]) {
                denora_free(c->excepts[i]);
            } else {
                alog(LOG_ERROR, langstr(ALOG_EXCEPTION_FREE_ERROR),
                     c->name, i);
            }
        }
        if (c->exceptsize) {
            denora_free(c->excepts);
        }
        if (denora->do_sql) {
            sql_channel_exception(ALL, c->name, NULL);
        }
    }

    SET_SEGV_LOCATION();

    if (ircd->invitemode) {
        for (i = 0; i < c->invitecount; ++i) {
            if (c->invite[i]) {
                denora_free(c->invite[i]);
            } else {
                alog(LOG_ERROR, langstr(ALOG_INVITE_FREE_ERROR), c->name,
                     i);
            }
        }
        if (c->invitesize) {
            denora_free(c->invite);
        }
        if (denora->do_sql) {
            sql_channel_invite(ALL, c->name, NULL);
        }
    }

    if (c->next) {
        c->next->prev = c->prev;
    }
    if (c->prev) {
        c->prev->next = c->next;
    } else {
        chanlist[HASH(c->name)] = c->next;
    }
    SET_SEGV_LOCATION();
    chans_hourly(NULL);
    denora_free(c);
}

/*************************************************************************/

void del_ban(Channel * chan, char *mask)
{
    char **s = chan->bans;
    int i = 0;

    SET_SEGV_LOCATION();

    /* Sanity check as it seems some IRCD will just send -b without a mask */
    if (!mask) {
        return;
    }

    while (i < chan->bancount && strcmp(*s, mask) != 0) {
        i++;
        s++;
    }

    if (i < chan->bancount) {
        chan->bancount--;
        if (i < chan->bancount) {
            memmove(s, s + 1, sizeof(char *) * (chan->bancount - i));
        }
        alog(LOG_DEBUG, langstr(ALOG_BAN_DEL_CHAN), mask, chan->name);
    }
    SET_SEGV_LOCATION();

    if (denora->do_sql) {
        sql_channel_ban(DEL, chan->name, mask);
    }
}

/*************************************************************************/

void del_exception(Channel * chan, char *mask)
{
    int i;
    int reset = 0;

    SET_SEGV_LOCATION();

    /* Sanity check as it seems some IRCD will just send -e without a mask */
    if (!mask) {
        return;
    }

    for (i = 0; i < chan->exceptcount; i++) {
        if ((!reset) && (stricmp(chan->excepts[i], mask) == 0)) {
            denora_free(chan->excepts[i]);
            reset = 1;
        }
        if (reset) {
            chan->excepts[i] =
                (i == chan->exceptcount) ? NULL : chan->excepts[i + 1];
        }
    }

    if (reset) {
        chan->exceptcount--;
    }
    SET_SEGV_LOCATION();

    alog(LOG_DEBUG, langstr(ALOG_EXCEPTION_DEL_CHAN), mask, chan->name);

    if (denora->do_sql) {
        sql_channel_exception(DEL, chan->name, mask);
    }
}

/*************************************************************************/

void del_invite(Channel * chan, char *mask)
{
    int i;
    int reset = 0;

    /* Sanity check as it seems some IRCD will just send -I without a mask */
    if (!mask) {
        return;
    }

    SET_SEGV_LOCATION();

    for (i = 0; i < chan->invitecount; i++) {
        if ((!reset) && (stricmp(chan->invite[i], mask) == 0)) {
            denora_free(chan->invite[i]);
            reset = 1;
        }
        if (reset) {
            chan->invite[i] =
                ((i == chan->invitecount) ? NULL : chan->invite[i + 1]);
        }
    }

    if (reset) {
        chan->invitecount--;
    }

    alog(LOG_DEBUG, langstr(ALOG_INVITE_DEL_CHAN), mask, chan->name);
    SET_SEGV_LOCATION();
    if (denora->do_sql) {
        sql_channel_invite(DEL, chan->name, mask);
    }
}

/*************************************************************************/

char *get_flood(Channel * chan)
{
    SET_SEGV_LOCATION();
    return chan->flood;
}

/*************************************************************************/

char *get_key(Channel * chan)
{
    SET_SEGV_LOCATION();
    return chan->key;
}

/*************************************************************************/

char *get_limit(Channel * chan)
{
    static char limit[16];

    SET_SEGV_LOCATION();

    if (chan->limit == 0) {
        return NULL;
    }

    snprintf(limit, sizeof(limit), "%lu", (unsigned long int) chan->limit);
    return limit;
}

/*************************************************************************/

char *get_redirect(Channel * chan)
{
    SET_SEGV_LOCATION();
    return chan->redirect;
}

/*************************************************************************/

Channel *join_user_update(User * user, Channel * chan, char *name)
{
    struct u_chanlist *c;

    SET_SEGV_LOCATION();

    /* If it's a new channel, so we need to create it first. */
    if (!chan) {
        chan = chan_create(name);
    }
    SET_SEGV_LOCATION();

    alog(LOG_DEBUG, langstr(ALOG_DEBUG_JOIN), user->nick, chan->name);

    c = scalloc(sizeof(*c), 1);
    c->next = user->chans;
    if (user->chans) {
        user->chans->prev = c;
    }
    user->chans = c;
    c->chan = chan;

    chan_adduser2(user, chan);

    SET_SEGV_LOCATION();

    return chan;
}

/*************************************************************************/

void set_flood(Channel * chan, char *value)
{
    SET_SEGV_LOCATION();

    if (chan->flood) {
        denora_free(chan->flood);
    }

    chan->flood = (value ? sstrdup(value) : NULL);

    alog(LOG_DEBUG, langstr(ALOG_DEBUG_FLOOD_MODE),
         chan->name, (chan->flood ? chan->flood : langstr(ALOG_NO_FLOOD)));
}

/*************************************************************************/

void set_key(Channel * chan, char *value)
{
    SET_SEGV_LOCATION();

    if (chan->key) {
        denora_free(chan->key);
    }

    chan->key = (value ? sstrdup(value) : NULL);

    alog(LOG_DEBUG, langstr(ALOG_KEY_SET_TO), chan->name,
         (chan->key ? chan->key : langstr(ALOG_NO_KEY)));
}

/*************************************************************************/

void set_limit(Channel * chan, char *value)
{
    SET_SEGV_LOCATION();

    chan->limit = (value ? strtoul(value, NULL, 10) : 0);

    alog(LOG_DEBUG, langstr(ALOG_LIMIT_SET_TO), chan->name, chan->limit);
}

/*************************************************************************/

void set_redirect(Channel * chan, char *value)
{
    SET_SEGV_LOCATION();

    if (chan->redirect) {
        denora_free(chan->redirect);
    }
    chan->redirect = (value ? sstrdup(value) : NULL);

    alog(LOG_DEBUG, langstr(ALOG_REDIRECT_SET_TO), chan->name,
         (chan->redirect ? chan->redirect : langstr(ALOG_NO_REDIRECT)));
}

/*************************************************************************/
