/* StatServ core functions
 *
 * (C) 2003-2005 Anope Team
 * Contact us at info@anope.org
 *
 * Please read COPYING and README for further details.
 *
 * Based on the original code of Epona by Lara.
 * Based on the original code of Services by Andy Church. 
 * 
 * $Id: ss_core.c 389 2005-07-05 22:45:48Z trystan $
 *
 */
/*************************************************************************/

#include "denora.h"

static int do_shutdown(User * u);
static int do_restart(User * u);
static int do_reload(User * u);
int DenoraInit(int argc, char **argv);
void DenoraFini(void);

/**
 * Create the command, and tell anope about it.
 * @param argc Argument count
 * @param argv Argument list
 * @return MOD_CONT to allow the module, MOD_STOP to stop it
 **/
int DenoraInit(int argc, char **argv)
{
    Command *c;

    if (denora->debug >= 2) {
        protocol_debug(NULL, argc, argv);
    }
    moduleAddAuthor("Denora");
    moduleAddVersion("$Id: ss_core.c 389 2005-07-05 22:45:48Z trystan $");
    moduleSetType(CORE);

    c = createCommand("SHUTDOWN",  do_shutdown,   is_stats_admin,      -1, -1,-1, STAT_HELP_SHUTDOWN);
    moduleAddCommand(STATSERV, c, MOD_UNIQUE);

    c = createCommand("RESTART",   do_restart,    is_stats_admin,      -1, -1,-1, STAT_HELP_RESTART);
    moduleAddCommand(STATSERV, c, MOD_UNIQUE);

    c = createCommand("RELOAD",    do_reload,     is_stats_admin,      -1, -1,-1, STAT_HELP_RELOAD);
    moduleAddCommand(STATSERV, c, MOD_UNIQUE);

    return MOD_CONT;
}

/**
 * Unload the module
 **/
void DenoraFini(void)
{

}

/* SHUTDOWN command. */

static int do_shutdown(User * u)
{
    SET_SEGV_LOCATION();

    denora->qmsg = calloc(32 + strlen(u->nick), 1);
    if (!denora->qmsg)
        denora->qmsg =
            sstrdup("SHUTDOWN command received, but out of memory!");
    else
        sprintf(denora->qmsg, "SHUTDOWN command received from %s",
                u->nick);

    SET_SEGV_LOCATION();

    denora->save_data = 1;
    denora->delayed_quit = 1;
    return MOD_CONT;
}

/*************************************************************************/

/* RESTART command */

static int do_restart(User * u)
{
#ifdef STATS_BIN
    denora->qmsg = calloc(31 + strlen(u->nick), 1);
    if (!denora->qmsg)
        denora->qmsg =
            sstrdup("RESTART command received, but out of memory!");
    else
        sprintf(denora->qmsg, "RESTART command received from %s", u->nick);

    /*    raise(SIGHUP); */
    do_restart_denora();
#else
    notice_lang(s_StatServ, u, STAT_CANNOT_RESTART);
#endif
    return MOD_CONT;
}

/*************************************************************************/

static int do_reload(User * u)
{
    if (initconf(denora->config, 1) == -1) {
        denora->qmsg = calloc(28 + strlen(u->nick), 1);
        if (!denora->qmsg)
            denora->qmsg =
                sstrdup
                ("Error during the reload of the configuration file, but out of memory!");
        else
            sprintf(denora->qmsg, "Error Reading Configuration File");
        denora->quitting = 1;
    } else {
        merge_confs();
    }

    notice_lang(s_StatServ, u, STAT_RELOAD);
    return MOD_CONT;
}

/*************************************************************************/

