/*
 *
 * (C) 2004-2005 Denora Team
 * Contact us at info@nomadirc.net
 *
 * Please read COPYING and README for furhter details.
 *
 * Based on the original code of Anope by Anope Team.
 * Based on the original code of Thales by Lucas.
 * 
 * $Id: db.c 427 2005-07-23 08:40:36Z trystan $
 *
 */

#include "denora.h"

/* last time we cleaned the server table */
static int ServerLastClean = -1;

/* last time we cleaned the user table */
static int UserLastClean = -1;

/*************************************************************************/

void load_stats_db(void)
{
    DenoraDBFile *dbptr = scalloc(1, sizeof(DenoraDBFile));
    char *key, *value;
    int retval = 0;

    fill_db_ptr(dbptr, 0, STATSDB_VERSION, s_StatServ, statsDB);
    SET_SEGV_LOCATION();

    /* let's remove existing temp files here, because we only load dbs on startup */
    remove(dbptr->temp_name);

    /* Open the db, fill the rest of dbptr and allocate memory for key and value */
    if (new_open_db_read(dbptr, &key, &value)) {
        SET_SEGV_LOCATION();
        free(dbptr);
        return;                 /* Bang, an error occurred */
    }

    while (1) {
        /* read a new entry and fill key and value with it -Certus */
        retval = new_read_db_entry(&key, &value, dbptr->fptr);

        if (retval == DB_READ_ERROR) {
            alog(LOG_NORMAL, langstr(ALOG_DB_ERROR), dbptr->filename);
            new_close_db(dbptr->fptr, &key, &value);
            SET_SEGV_LOCATION();
            free(dbptr);
            return;
        } else if (retval == DB_EOF_ERROR) {
            alog(LOG_EXTRADEBUG, langstr(ALOG_DEBUG_DB_OK),
                 dbptr->filename);
            new_close_db(dbptr->fptr, &key, &value);
            SET_SEGV_LOCATION();
            free(dbptr);
            return;
        } else if (retval == DB_READ_BLOCKEND) {        /* DB_READ_BLOCKEND */
            /* a channel has completely been read. put any checks in here! */
        } else {
            /* DB_READ_SUCCESS */

            if (!*value || !*key) {
                continue;
            }
            SET_SEGV_LOCATION();

            if (!stricmp(key, "usermax")) {
                stats->users_max = atoi(value);
            } else if (!stricmp(key, "usermaxtime")) {
                stats->users_max_time = atoi(value);
            } else if (!stricmp(key, "channelmax")) {
                stats->chans_max = atoi(value);
            } else if (!stricmp(key, "channelmaxtime")) {
                stats->chans_max_time = atoi(value);
            } else if (!stricmp(key, "servermax")) {
                stats->servers_max = atoi(value);
            } else if (!stricmp(key, "servermaxtime")) {
                stats->servers_max_time = atoi(value);
            } else if (!stricmp(key, "opermax")) {
                stats->opers_max = atoi(value);
            } else if (!stricmp(key, "opermaxtime")) {
                stats->opers_max_time = atoi(value);
            }
        }                       /* else */
    }                           /* while */
}

/*************************************************************************/

void save_stats_db(void)
{
    DenoraDBFile *dbptr = scalloc(1, sizeof(DenoraDBFile));

    fill_db_ptr(dbptr, 0, STATSDB_VERSION, s_StatServ, statsDB);
    SET_SEGV_LOCATION();

    /* time to backup the old db */
    rename(statsDB, dbptr->temp_name);

    if (new_open_db_write(dbptr)) {
        rename(dbptr->temp_name, statsDB);
        free(dbptr);
        return;                 /* Bang, an error occurred */
    }
    SET_SEGV_LOCATION();
    new_write_db_entry("usermax", dbptr, "%ld", stats->users_max);
    new_write_db_entry("usermaxtime", dbptr, "%ld",
                       (long int) stats->users_max_time);
    new_write_db_entry("channelmax", dbptr, "%ld", stats->chans_max);
    new_write_db_entry("channelmaxtime", dbptr, "%ld",
                       (long int) stats->chans_max_time);
    new_write_db_entry("servermax", dbptr, "%ld", stats->servers_max);
    new_write_db_entry("servermaxtime", dbptr, "%ld",
                       (long int) stats->servers_max_time);
    new_write_db_entry("opermax", dbptr, "%ld", stats->opers_max);
    new_write_db_entry("opermaxtime", dbptr, "%ld",
                       (long int) stats->opers_max_time);
    new_write_db_endofblock(dbptr);

    SET_SEGV_LOCATION();
    if (dbptr) {
        new_close_db(dbptr->fptr, NULL, NULL);  /* close file */
        remove(dbptr->temp_name);       /* saved successfully, no need to keep the old one */
        free(dbptr);            /* free the db struct */
    }
}

/*************************************************************************/

void db_connect(void)
{
    Exclude *e, *next;
    ChannelStats *cs;

    if (!denora->do_sql) {
        alog(LOG_ERROR, langstring(ALOG_FAILED_SQL_CONNECT),
             rdb_error_msg());
        return;
    }
    SET_SEGV_LOCATION();

    /* Cleaning up the database */
    rdb_clear_table(ChanTable);
    rdb_clear_table(IsOnTable);
    rdb_clear_table(ServerTable);
    rdb_clear_table(UserTable);
    rdb_clear_table(GlineTable);
    rdb_clear_table(ChanBansTable);
    rdb_clear_table(CTCPTable);

    if (ircd->spamfilter) {
        rdb_clear_table(SpamTable);
    }
    if (ircd->except) {
        rdb_clear_table(ChanExceptTable);
    }
    if (ircd->invitemode) {
        rdb_clear_table(ChanInviteTable);
    }
    rdb_clear_table(TLDTable);
    if (ircd->sgline_table) {
        rdb_clear_table(SglineTable);
    }
    if (ircd->sqline_table) {
        rdb_clear_table(SqlineTable);
    }

    e = first_exclude();
    while (e) {
        next = next_exclude();
        rdb_query("DELETE FROM %s WHERE uname=\"%s\"", UStatsTable,
                  e->name);
        e = next;
    }
    rdb_query("DELETE FROM %s WHERE uname=\"%s\"", UStatsTable,
              s_StatServ);
    if (s_StatServ_alias) {
        rdb_query("DELETE FROM %s WHERE uname=\"%s\"",
                  UStatsTable, s_StatServ_alias);
    }

    if (LogChannel) {
        if (!(cs = find_cs(LogChannel))) {
            del_cs(cs);
            save_cs_db();
            rdb_query("DELETE FROM %s WHERE chan=\"%s\"", CStatsTable,
                      LogChannel);
            rdb_query("DELETE FROM %s WHERE chan=\"%s\"", UStatsTable,
                      LogChannel);
        }
    }
    SET_SEGV_LOCATION();
}

/*************************************************************************/

/* serv should be db_escape'd before call */
/* -1 if server not found, servid else */
int db_checkserver(char *serv)
{
    int servid = -1;
    Server *s;
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    s = findserver(serv);
    if (s && s->sqlid) {
        return s->sqlid;
    }

    SET_SEGV_LOCATION();

    if (!denora->do_sql) {
        return -1;
    }

    rdb_query("SELECT servid FROM %s WHERE server=\"%s\"", ServerTable,
              serv);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    if (mysql_res) {
        if (mysql_num_rows(mysql_res)) {
            servid = atoi(*mysql_fetch_row(mysql_res));
        }
        SET_SEGV_LOCATION();
        mysql_free_result(mysql_res);
    }
#endif
    return servid;
}

/*************************************************************************/

/* serv should be db_escape'd before call */
int db_getserver(char *serv)
{
    Server *s;
    int res = 0;
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    SET_SEGV_LOCATION();

    s = findserver(serv);
    if (s && s->sqlid) {
        return s->sqlid;
    }

    if (!denora->do_sql) {
        return res;
    }

    if (!serv) {
        return res;
    }

    rdb_query("SELECT servid FROM %s WHERE server=\'%s\'", ServerTable,
              serv);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    if (mysql_res) {
        if (mysql_num_rows(mysql_res)) {
            res = atoi(*mysql_fetch_row(mysql_res));
        }
        SET_SEGV_LOCATION();
        mysql_free_result(mysql_res);
    }
#endif
    return res;
}

/*************************************************************************/

/* nick should be db_escape'd before call */
/* -1 if nick not found, nickid else */
int db_checknick(char *nick)
{
    int nickid = -1;
    User *u;
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    SET_SEGV_LOCATION();

    u = finduser(nick);
    if (u && u->sqlid) {
        return u->sqlid;
    }

    if (!denora->do_sql) {
        return -1;
    }

    rdb_query("SELECT nickid FROM %s WHERE nick=\"%s\"", UserTable, nick);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    if (mysql_res) {
        if (mysql_num_rows(mysql_res)) {
            nickid = atoi(*mysql_fetch_row(mysql_res));
        }
        SET_SEGV_LOCATION();
        mysql_free_result(mysql_res);
    } else {
        nickid = -1;
    }
#endif
    return nickid;
}

/*************************************************************************/

/* nick should be db_escape'd before call */
int db_getnick_unsure(char *nick)
{
    User *u;
    int res = 0;
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    SET_SEGV_LOCATION();

    u = finduser(nick);
    if (u && u->sqlid) {
        return u->sqlid;
    }

    if (!denora->do_sql) {
        return -1;
    }

    rdb_query("SELECT nickid FROM %s WHERE nick=\'%s\'", UserTable, nick);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    if (mysql_res) {
        if (mysql_num_rows(mysql_res)) {
            res = atoi(*mysql_fetch_row(mysql_res));
        } else {
            res = -1;
        }
        SET_SEGV_LOCATION();
        mysql_free_result(mysql_res);
    }
#endif
    return res;
}

/*************************************************************************/

/* nick should be db_escape'd before call */
int db_getnick(char *nick)
{
    User *u;
    int res = 0;
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    u = finduser(nick);
    if (u && u->sqlid) {
        return u->sqlid;
    }

    if (!denora->do_sql) {
        return 0;
    }
    SET_SEGV_LOCATION();

    rdb_query("SELECT nickid FROM %s WHERE nick=\'%s\'", UserTable, nick);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    if (mysql_res) {
        if (mysql_num_rows(mysql_res)) {
            res = atoi(*mysql_fetch_row(mysql_res));
        } else {
            alog(LOG_NONEXISTANT, "nickname not found ! %s", nick);
        }
        mysql_free_result(mysql_res);
    }
#endif
    return res;
}

/*************************************************************************/

int db_getservfromnick(char *nick)
{
    int res = 0;
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    if (!denora->do_sql) {
        return -1;
    }
    SET_SEGV_LOCATION();

    rdb_query("SELECT servid FROM %s WHERE nick=\'%s\'", UserTable, nick);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    if (mysql_res) {
        if (mysql_num_rows(mysql_res)) {
            res = atoi(*mysql_fetch_row(mysql_res));
        } else {
            alog(LOG_NONEXISTANT, "nickname not found ! %s", nick);
        }
        SET_SEGV_LOCATION();
        mysql_free_result(mysql_res);
    }
#endif
    return res;
}

/*************************************************************************/

void db_removenick(char *nick)
{
    int nickid = db_getnick(nick);
    /* TODO */

    SET_SEGV_LOCATION();

    if (!denora->do_sql) {
        return;
    }

    if (nickid == 0) {
        alog(LOG_NONEXISTANT, "nickid 0");
        return;
    }
    SET_SEGV_LOCATION();

    db_removefromchans(nickid);
    if (UserCacheTime) {
        rdb_query
            ("UPDATE %s SET online=\"N\", lastquit=NOW(), servid=NULL WHERE nickid=%d",
             UserTable, nickid);
    } else {
        rdb_query("DELETE FROM %s WHERE nickid=%d", UserTable, nickid);
    }
    SET_SEGV_LOCATION();
}

/*************************************************************************/

void db_removefromchans(int nickid)
{
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif
    char **res;
    int users;
    int chanid;

    SET_SEGV_LOCATION();

    if (!denora->do_sql) {
        return;
    }

    rdb_query
        ("SELECT %s.chanid, channel FROM %s, %s WHERE nickid=%d AND %s.chanid = %s.chanid",
         IsOnTable, IsOnTable, ChanTable, nickid, ChanTable, IsOnTable);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
#endif
    rdb_query("DELETE FROM %s WHERE nickid=%d", IsOnTable, nickid);
    SET_SEGV_LOCATION();
#ifdef USE_MYSQL
    if (mysql_res) {
        while ((res = mysql_fetch_row(mysql_res))) {
            char *chan = rdb_escape(res[1]);
            users = db_getchannel_users(chan);
            chanid = db_getchannel(chan);
            if (users) {
                users--;
                rdb_query("UPDATE %s SET currentusers=%d WHERE chanid=%d",
                          ChanTable, users, chanid);
            }
            db_checkemptychan(atoi(res[0]));
            denora_free(chan);
        }
        SET_SEGV_LOCATION();
        mysql_free_result(mysql_res);
    }
#endif
}

/*************************************************************************/

void db_checkemptychan(int chanid)
{
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    SET_SEGV_LOCATION();

    if (!denora->do_sql) {
        return;
    }
    rdb_query("SELECT chanid FROM %s WHERE chanid=%d", IsOnTable, chanid);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    if (mysql_res) {
        SET_SEGV_LOCATION();
        if (!mysql_num_rows(mysql_res)) {
            rdb_query("DELETE FROM %s WHERE chanid=%d", ChanTable, chanid);
        }
        SET_SEGV_LOCATION();
        mysql_free_result(mysql_res);
    }
#endif
}

/*************************************************************************/

int db_getlusers(int type)
{
    int retcode = 0;
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    if (!denora->do_sql) {
        return 0;
    }
    SET_SEGV_LOCATION();

    switch (type) {
    case LUSERS_USERS:
        rdb_query("SELECT COUNT(*) FROM %s WHERE mode_li=\'N\'",
                  UserTable);
        break;
    case LUSERS_USERSINV:
        rdb_query("SELECT COUNT(*) FROM %s WHERE mode_li=\'Y\'",
                  UserTable);
        break;
    case LUSERS_OPERS:
        rdb_query("SELECT COUNT(*) FROM %s WHERE mode_lo=\'Y\'",
                  UserTable);
        break;
    case LUSERS_CHAN:
        rdb_query("SELECT COUNT(*) FROM %s", ChanTable);
        break;
    case LUSERS_SERV:
        rdb_query("SELECT COUNT(*) FROM %s", ServerTable);
        break;
    case LUSERS_USERSGLOB:
        rdb_query("SELECT COUNT(*) FROM %s", UserTable);
        break;
    case LUSERS_USERSMAX:
        rdb_query("SELECT val FROM %s WHERE type='users'", MaxValueTable);
        break;
    }
    SET_SEGV_LOCATION();
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    if (mysql_res) {
        if (mysql_num_rows(mysql_res)) {
            retcode = atoi(*mysql_fetch_row(mysql_res));
        } else {
            retcode = -1;
        }
        SET_SEGV_LOCATION();
        mysql_free_result(mysql_res);
    }
#endif
    return retcode;
}

/*************************************************************************/

/* chan should be db_escape'd before call */
int db_getchannel(char *chan)
{
    int res = 0;
    Channel *c;
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    strtolwr(chan);
    SET_SEGV_LOCATION();

    c = findchan(chan);
    if (c && c->sqlid) {
        return c->sqlid;
    }

    if (!denora->do_sql) {
        return -1;
    }

    rdb_query("SELECT chanid FROM %s WHERE channel=\'%s\'", ChanTable,
              chan);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    if (mysql_res) {
        if (mysql_num_rows(mysql_res)) {
            res = atoi(*mysql_fetch_row(mysql_res));
        } else {
            alog(LOG_NONEXISTANT, "channel not found ! %s", chan);
        }
        SET_SEGV_LOCATION();
        mysql_free_result(mysql_res);
    }
#endif
    return res;
}

/*************************************************************************/

char *db_getchannel_byid(int chanid)
{
    char *res = NULL;
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
    MYSQL_ROW row;
#endif

    if (!denora->do_sql) {
        return NULL;
    }
    SET_SEGV_LOCATION();

    rdb_query("SELECT channel FROM %s WHERE chanid=%d", ChanTable, chanid);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    if (mysql_res) {
        if (!mysql_num_rows(mysql_res)) {
            alog(LOG_NONEXISTANT, "channel not found ! %d", chanid);
        } else {
            row = mysql_fetch_row(mysql_res);
            res = rdb_escape(row[0]);
        }
        SET_SEGV_LOCATION();
        mysql_free_result(mysql_res);
    }
#endif
    return res;
}

/*************************************************************************/

/* chan should be db_escape'd before call */
int db_getchannel_users(char *chan)
{
    int res = 0;
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    SET_SEGV_LOCATION();
    strtolwr(chan);

    if (!denora->do_sql) {
        return -1;
    }
    SET_SEGV_LOCATION();

    rdb_query("SELECT currentusers FROM %s WHERE channel=\'%s\'",
              ChanTable, chan);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    SET_SEGV_LOCATION();
    if (mysql_res) {
        if (mysql_num_rows(mysql_res)) {
            res = strtol(*mysql_fetch_row(mysql_res), NULL, 10);
        } else {
            alog(LOG_DEBUG,
                 "debug: unable to find the requested channel %s", chan);
        }
        SET_SEGV_LOCATION();
        mysql_free_result(mysql_res);
    }
#endif
    return res;
}

/*************************************************************************/

/* chan should be db_escape'd before call */
/* chan is created if not exists */
int db_getchancreate(char *chan)
{
    int res = -1;
    Channel *c;
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    SET_SEGV_LOCATION();
    strtolwr(chan);

    c = findchan(chan);
    if (c && c->sqlid) {
        return c->sqlid;
    }

    if (!denora->do_sql) {
        return -1;
    }
    rdb_query("SELECT chanid FROM %s WHERE channel=\'%s\'",
              ChanTable, chan);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    if (mysql_res) {
        if (mysql_num_rows(mysql_res))
            res = atoi(*mysql_fetch_row(mysql_res));
        mysql_free_result(mysql_res);
    }
#endif
    SET_SEGV_LOCATION();
    if (res == -1) {
        rdb_query("INSERT INTO %s (channel) VALUES (\'%s\')", ChanTable,
                  chan);
        res = rdb_insertid();
    }
    SET_SEGV_LOCATION();
    if (c && res) {
        c->sqlid = res;
    }
    return res;
}

/*************************************************************************/

/* cleanup the server table, removing old entries */
void db_cleanserver()
{
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif
    int curtime = time(NULL);

    SET_SEGV_LOCATION();

    if (!denora->do_sql) {
        return;
    }

    if (ServerLastClean == -1) {
        ServerLastClean = curtime;
    }
    SET_SEGV_LOCATION();
    if (curtime > (ServerLastClean + ServerCleanFreq)) {
        ServerLastClean = curtime;
        rdb_query
            ("SELECT server FROM %s WHERE online=\"N\" AND lastsplit<FROM_UNIXTIME(%d)",
             ServerTable, curtime - ServerCacheTime);
#ifdef USE_MYSQL
        mysql_res = mysql_store_result(mysql);
        if (mysql_num_rows(mysql_res)) {
            mysql_free_result(mysql_res);
            rdb_query
                ("DELETE FROM %s WHERE online=\"N\" AND lastsplit<FROM_UNIXTIME(%d)",
                 ServerTable, curtime - ServerCacheTime);
        } else {
            if (mysql_res) {
                mysql_free_result(mysql_res);
            }
        }
#endif
        SET_SEGV_LOCATION();
    }
}

/*************************************************************************/

/* cleanup the user table, removing old entries */
void db_cleanuser()
{
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif
    int curtime = time(NULL);

    if (!denora->do_sql) {
        return;
    }

    if (UserLastClean == -1) {
        UserLastClean = curtime;
    }
    SET_SEGV_LOCATION();

    if (curtime > (UserLastClean + UserCleanFreq)) {
        UserLastClean = curtime;
        rdb_query
            ("SELECT nick FROM %s WHERE online=\"N\" AND lastquit<FROM_UNIXTIME(%d)",
             UserTable, curtime - UserCacheTime);
#ifdef USE_MYSQL
        mysql_res = mysql_store_result(mysql);
        if (mysql_res) {
            if (mysql_num_rows(mysql_res)) {
                rdb_query
                    ("DELETE FROM %s WHERE online=\"N\" AND lastquit<FROM_UNIXTIME(%d)",
                     UserTable, curtime - UserCacheTime);
            }
            mysql_free_result(mysql_res);
        }
#endif
        SET_SEGV_LOCATION();
    }
}

/*************************************************************************/

/* check if chans > chans_max */
void do_checkchansmax()
{
    if (stats->chans > stats->chans_max) {
        stats->chans_max = stats->chans;
        stats->chans_max_time = time(NULL);
    }
    if (!stats->chans_max_time) {
        stats->chans_max_time = time(NULL);
    }
    SET_SEGV_LOCATION();

    if (stats->chans > stats->daily_chans) {
        stats->daily_chans++;
        stats->daily_chans_time = time(NULL);
    }
    rdb_query("UPDATE %s SET val=%d, time=%ld WHERE type='chans'",
              CurrentTable, stats->chans, time(NULL));
    rdb_query("UPDATE %s SET val=%d, time=%ld WHERE type='channels'",
              MaxValueTable, stats->chans_max,
              (long int) stats->chans_max_time);

}

/*************************************************************************/

/* check if users > users_max */
void add_current_user(Server * s)
{
    SET_SEGV_LOCATION();

    if (s) {
        s->ss->currentusers++;
        s->ss->totalusersever++;
    }
    stats->users++;
    stats->totalusersever++;

    SET_SEGV_LOCATION();
    if (stats->users > stats->users_max) {
        stats->users_max = stats->users;
        stats->users_max_time = time(NULL);
    }
    if (!stats->users_max_time) {
        stats->users_max_time = time(NULL);
    }
    SET_SEGV_LOCATION();
    if (stats->users > stats->daily_users) {
        stats->daily_users++;
        stats->daily_users_time = time(NULL);
    }
    SET_SEGV_LOCATION();
    if (s) {
        if (s->ss->currentusers > s->ss->maxusers) {
            s->ss->maxusers = s->ss->currentusers;
            s->ss->maxusertime = time(NULL);
        }
    }
    SET_SEGV_LOCATION();
    if (denora->do_sql) {
        rdb_query("UPDATE %s SET val=%d, time=%ld WHERE type='users'",
                  CurrentTable, stats->users, time(NULL));
        rdb_query
            ("UPDATE %s SET val=%d, time=%ld WHERE type='daily_users'",
             CurrentTable, stats->daily_users, time(NULL));
        rdb_query
            ("UPDATE %s SET val=%d, time=%ld WHERE type='users'",
             MaxValueTable, stats->users_max,
             (long int) stats->users_max_time);
        if (s) {
            rdb_query
                ("UPDATE %s SET currentusers=%d, maxusers=%d, maxusertime=%d WHERE server='%s'",
                 ServerTable, s->ss->currentusers, s->ss->maxusers,
                 s->ss->maxusertime, s->name);
        }
    }
    users_hourly(NULL);
}

/*************************************************************************/

void del_current_user(Server * s)
{
    SET_SEGV_LOCATION();
    if (s) {
        s->ss->currentusers--;
    }
    stats->users--;
    SET_SEGV_LOCATION();

    rdb_query("UPDATE %s SET val=%d, time=%ld WHERE type='users'",
              CurrentTable, stats->users, time(NULL));
    if (s) {
        rdb_query("UPDATE %s SET currentusers=%d WHERE server='%s'",
                  ServerTable, s->ss->currentusers, s->name);
    }
    users_hourly(NULL);
}

/*************************************************************************/

/* check if servs > servs_max */
void do_checkservsmax()
{
    SET_SEGV_LOCATION();

    if (stats->servers > stats->servers_max) {
        stats->servers_max = stats->servers;
        stats->servers_max_time = time(NULL);
    }
    if (!stats->servers_max_time) {
        stats->servers_max_time = time(NULL);
    }
    SET_SEGV_LOCATION();

    if (stats->servers > stats->daily_servers) {
        stats->daily_servers++;
        stats->daily_servers_time = time(NULL);
    }
    rdb_query("UPDATE %s SET val=%d, time=%ld WHERE type='servers'",
              CurrentTable, stats->servers, time(NULL));
    rdb_query("UPDATE %s SET val=%d, time=%ld WHERE type='servers'",
              MaxValueTable, stats->servers_max,
              (long int) stats->servers_max_time);

}

/*************************************************************************/

void add_oper_count(User * u)
{
    Server *s;

    stats->opers++;
    s = u->server;

    SET_SEGV_LOCATION();

    if (s) {
        s->ss->opers++;
    }
    SET_SEGV_LOCATION();

    if (stats->opers > stats->opers_max) {
        stats->opers_max = stats->opers;
        stats->opers_max_time = time(NULL);
    }

    SET_SEGV_LOCATION();

    if (stats->opers > stats->daily_opers) {
        stats->daily_opers++;
        stats->daily_opers_time = time(NULL);
    }
    SET_SEGV_LOCATION();

    if (s) {
        if (s->ss->opers > s->ss->maxopers) {
            s->ss->maxopers = s->ss->opers;
            s->ss->maxopertime = time(NULL);
        }
        if (!s->ss->maxopertime) {
            s->ss->maxopertime = time(NULL);
        }
        rdb_query
            ("UPDATE %s SET opers=%d, maxopers=%ld, maxopertime=%ld WHERE server='%s'",
             ServerTable, s->ss->opers, s->ss->maxopers,
             (long int) s->ss->maxopertime, s->name);

    }
    rdb_query("UPDATE %s SET val=%d, time=%ld WHERE type='opers'",
              CurrentTable, stats->opers, (long int) time(NULL));
    rdb_query("UPDATE %s SET val=%d, time=%ld WHERE type='opers'",
              MaxValueTable, stats->opers_max,
              (long int) stats->opers_max_time);

}

/*************************************************************************/

void del_oper_count(User * u)
{
    SET_SEGV_LOCATION();

    stats->opers--;
    if (u && u->server && u->server->ss) {
        u->server->ss->opers--;
        rdb_query
            ("UPDATE %s SET opers=%d, maxopers=%ld, maxopertime=%ld WHERE server='%s'",
             ServerTable, u->server->ss->opers, u->server->ss->maxopers,
             (long int) u->server->ss->maxopertime, u->server->name);
    }

    SET_SEGV_LOCATION();

    rdb_query("UPDATE %s SET val=%d, time=%ld WHERE type='opers'",
              CurrentTable, stats->opers, time(NULL));
}

/*************************************************************************/

int users_daily(const char *name)
{
    SET_SEGV_LOCATION();

    if (name) {
        stats->daily_users = stats->users;
        stats->daily_users_time = time(NULL);
        stats->daily_opers = stats->opers;
        stats->daily_opers_time = time(NULL);
        stats->daily_servers = stats->servers;
        stats->daily_servers_time = time(NULL);
        stats->daily_chans = stats->chans;
        stats->daily_chans_time = time(NULL);
    }

    SET_SEGV_LOCATION();

    return MOD_CONT;
}

/*************************************************************************/

int chans_hourly(const char *name)
{
    static struct tm *mytime;
    time_t ts;
    int year;
    int month;
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    ts = time(NULL);
    mytime = localtime(&ts);

    year = 1900 + mytime->tm_year;
    month = 1 + mytime->tm_mon;

    if (name) {
        alog(LOG_DEBUG, "Executing for %s", name);
    }

    if (!denora->do_sql) {
        return MOD_CONT;
    }

    rdb_query("SELECT * FROM %s WHERE day=%d and month=%d and year=%d",
              ChanStatsTable, mytime->tm_mday, month, year);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    if (mysql_res) {
        if (mysql_num_rows(mysql_res) > 0) {
            rdb_query
                ("UPDATE %s SET time_%d=%d WHERE day=%d and month=%d and year=%d",
                 ChanStatsTable, mytime->tm_hour, stats->chans,
                 mytime->tm_mday, month, year);
        } else {
            rdb_query
                ("INSERT INTO %s (day, month, year, time_%d) VALUES (%d, %d, %d, %d)",
                 ChanStatsTable, mytime->tm_hour, mytime->tm_mday, month,
                 year, stats->chans);
        }
        mysql_free_result(mysql_res);
    }
    SET_SEGV_LOCATION();
#endif
    SET_SEGV_LOCATION();
    return MOD_CONT;
}

/*************************************************************************/

int servers_hourly(const char *name)
{
    static struct tm *mytime;
    time_t ts;
    int year;
    int month;
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    ts = time(NULL);
    mytime = localtime(&ts);

    year = 1900 + mytime->tm_year;
    month = 1 + mytime->tm_mon;

    if (name) {
        alog(LOG_DEBUG, "Executing for %s", name);
    }

    if (!denora->do_sql) {
        return MOD_CONT;
    }

    rdb_query("SELECT * FROM %s WHERE day=%d and month=%d and year=%d",
              ServerStatsTable, mytime->tm_mday, month, year);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    if (mysql_res) {
        if (mysql_num_rows(mysql_res) > 0) {
            rdb_query
                ("UPDATE %s SET time_%d=%d WHERE day=%d and month=%d and year=%d",
                 ServerStatsTable, mytime->tm_hour, stats->servers,
                 mytime->tm_mday, month, year);
        } else {
            rdb_query
                ("INSERT INTO %s (day, month, year, time_%d) VALUES (%d, %d, %d, %d)",
                 ServerStatsTable, mytime->tm_hour, mytime->tm_mday, month,
                 year, stats->servers);
        }
        mysql_free_result(mysql_res);
    }
    SET_SEGV_LOCATION();
#endif
    SET_SEGV_LOCATION();
    return MOD_CONT;
}

/*************************************************************************/

int users_hourly(const char *name)
{
    static struct tm *mytime;
    time_t ts;
    int year;
    int month;
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    ts = time(NULL);
    mytime = localtime(&ts);

    year = 1900 + mytime->tm_year;
    month = 1 + mytime->tm_mon;

    if (name) {
        alog(LOG_DEBUG, "Executing for %s", name);
    }

    if (!denora->do_sql) {
        return MOD_CONT;
    }

    rdb_query("SELECT * FROM %s WHERE day=%d and month=%d and year=%d",
              StatsTable, mytime->tm_mday, month, year);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    if (mysql_res) {
        if (mysql_num_rows(mysql_res) > 0) {
            rdb_query
                ("UPDATE %s SET time_%d=%d WHERE day=%d and month=%d and year=%d",
                 StatsTable, mytime->tm_hour, stats->users,
                 mytime->tm_mday, month, year);
        } else {
            rdb_query
                ("INSERT INTO %s (day, month, year, time_%d) VALUES (%d, %d, %d, %d)",
                 StatsTable, mytime->tm_hour, mytime->tm_mday, month, year,
                 stats->users);
        }
        mysql_free_result(mysql_res);
    }
    SET_SEGV_LOCATION();
#endif
    SET_SEGV_LOCATION();
    return MOD_CONT;
}
