/* InspIRCd beta 4 functions
 *
 * (C) 2004-2005 Denora Team
 * Contact us at info@nomadirc.net
 *
 * Please read COPYING and README for furhter details.
 *
 * Based on the original code of Anope by Anope Team.
 * Based on the original code of Thales by Lucas.
 *
 * $Id: inspircd.c 433 2005-07-26 16:58:24Z drstein $
 *
 */

/*************************************************************************/

#include "denora.h"
#include "inspircd.h"

#ifdef __STRICT_ANSI__
#include <arpa/inet.h>
int inet_aton(const char *cp, struct in_addr *inp);
#endif

char* CreateSum(void);

#ifdef _WIN32 
#include "winsock.h" 
int inet_aton (const char *name, struct in_addr *addr) 
{ 
   uint32 a = inet_addr (name); 
   addr->s_addr = a; 
   return a != (uint32)-1; 
} 
#endif 

const char *xsumtable = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";

char* CreateSum(void) 
{ 
  int q = 0; 
  static char sum[8]; 
  sum[7] = '\0'; 
  for(q = 0; q < 7; q++) 
     sum[q] = xsumtable[rand()%52]; 
  return sum; 
} 

IRCDVar myIrcd[] = {
    {"InspIRCd 1.0 Beta 5",       /* ircd name */
     "+o",                      /* operserv mode */
     1,                         /* Vhost  */
     0,                         /* Supports SGlines     */
     0,                         /* sgline sql table */
     1,                         /* Supports SQlines     */
     0,                         /* sqline sql table */
     1,                         /* Supports SZlines     */
     0,                         /* Has exceptions +e    */
     0,                         /* vidents              */
     0,                         /* NICKIP               */
     1,                         /* VHOST ON NICK        */
     0,                         /* +f                   */
     1,                         /* +L                   */
     0,                         /* +f Mode                          */
     IRCD_CMODE_L,                   /* +L Mode                          */
     NULL,                      /* CAPAB Chan Modes             */
     1,                         /* We support TOKENS */
     0,                         /* TOKENS are CASE Sensitive */
     0,                         /* TIME STAMPS are BASE64 */
     0,                         /* +I support */
     0,                         /* SJOIN ban char */
     0,                         /* SJOIN except char */
     IRC_UMODE_x,                   /* umode for vhost */
     1,                         /* owner */
     1,                         /* protect */
     1,                         /* halfop */
     "aghiorswxA",
     "abcfhkilmnopqrstuvzACGHKLNOQRSV",
     0,
     0,
     'x',
     0,
     0,
     0,
     0,
     0,
     0,
     'b',
     0,
     0,
     0,
     0,
     NULL,
     1},
};


IRCDCAPAB myIrcdcap[] = {
    {
     CAPAB_NOQUIT,              /* NOQUIT       */
     0,                         /* TSMODE       */
     0,                         /* UNCONNECT    */
     0,                         /* NICKIP       */
     0,                         /* SJOIN        */
     0,                         /* ZIP          */
     0,                         /* BURST        */
     0,                         /* TS5          */
     0,                         /* TS3          */
     0,                         /* DKEY         */
     0,                         /* PT4          */
     0,                         /* SCS          */
     0,                         /* QS           */
     0,                         /* UID          */
     0,                         /* KNOCK        */
     0,                         /* CLIENT       */
     0,                         /* IPV6         */
     0,                         /* SSJ5         */
     0,                         /* SN2          */
     CAPAB_TOKEN,               /* TOKEN        */
     0,                         /* VHOST        */
     0,                         /* SSJ3         */
     0,                         /* NICK2        */
     0,                         /* UMODE2       */
     0,                         /* VL           */
     0,                         /* TLKEXT       */
     0,                         /* DODKEY       */
     0,                         /* DOZIP        */
     0,
     0, 0}
};

unsigned long myumodes[128] = {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0,
    0,
    0,
    0, 0, 0, 0, 0, 0, 0,
    0,
    0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0,
    IRC_UMODE_g,
    IRC_UMODE_h, IRC_UMODE_i, 0, 0, 0, 0, 0, IRC_UMODE_o,
    0,
    0, IRC_UMODE_r, IRC_UMODE_s, 0, 0, 0, IRC_UMODE_w,
    IRC_UMODE_x,
    0,
    0,
    0, 0, 0, 0, 0
};


char myCsmodes[128] = {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    0,
    0,
    0, 0, 0,
    'h',                        /* (37) % Channel halfops */
    0,
    0, 0, 0,
    'q',
    'v', 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    'o', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 'a', 0
};

CMMode myCmmodes[128] = {
    {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL},
    {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL},
    {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL},
    {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL},
    {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL},
    {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL},
    {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL},
    {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL},
    {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL},
    {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL},
    {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL},
    {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL},
    {NULL, NULL},
    {NULL, NULL},
    {add_ban, del_ban},
    {NULL, NULL},
    {NULL, NULL},
    {add_exception, del_exception},
    {NULL, NULL},
    {NULL, NULL},
    {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL},
    {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL},
    {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL},
};

CBMode myCbmodes[128] = {
    {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
    {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
    {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
    {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
    {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
    {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
    {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
    {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
    {0, 0},
    {IRCD_CMODE_A, NULL},
    {0, 0},                        /* B */
    {IRCD_CMODE_C, NULL},   /* C */
    {0, 0},                        /* D */
    {0, 0},                        /* E */
    {0, 0},                        /* F */
    {IRCD_CMODE_G, NULL},   /* G */
    {IRCD_CMODE_H, NULL},
    {0, 0},                        /* I */
    {0, 0},                        /* J */
    {IRCD_CMODE_K, NULL},   /* K */
    {IRCD_CMODE_L, set_redirect},
    {0, 0},                        /* M */
    {IRCD_CMODE_N, NULL},   /* N */
    {IRCD_CMODE_O, NULL},
    {0, 0},                        /* P */
    {IRCD_CMODE_Q, NULL},   /* Q */
    {IRCD_CMODE_R, NULL},   /* R */
    {IRCD_CMODE_S, NULL},   /* S */
    {0, 0},                        /* T */
    {0, 0},                        /* U */
    {IRCD_CMODE_V, NULL},   /* V */
    {0, 0},                        /* W */
    {0, 0},                        /* X */
    {0, 0},                        /* Y */
    {0, 0},                        /* Z */
    {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
    {0, 0},                        /* a */
    {0, 0},                        /* b */
    {IRCD_CMODE_c, NULL},
    {0, 0},                        /* d */
    {0, 0},                        /* e */
    {IRCD_CMODE_f, set_flood},
    {0, 0},                        /* g */
    {0, 0},                        /* h */
    {IRCD_CMODE_i, NULL},
    {0, 0},                        /* j */
    {IRCD_CMODE_k, set_key},
    {IRCD_CMODE_l, set_limit},
    {IRCD_CMODE_m, NULL},
    {IRCD_CMODE_n, NULL},
    {0, 0},                        /* o */
    {IRCD_CMODE_p, NULL},
    {0, 0},                        /* q */
    {IRCD_CMODE_r, NULL},
    {IRCD_CMODE_s, NULL},
    {IRCD_CMODE_t, NULL},
    {IRCD_CMODE_u, NULL},
    {0, 0},                        /* v */
    {0, 0},                        /* w */
    {0, 0},                        /* x */
    {0, 0},                        /* y */
    {IRCD_CMODE_z, NULL},
    {0, 0}, {0, 0}, {0, 0}, {0, 0}
};

CBModeInfo myCbmodeinfos[] = {
    {'c', IRCD_CMODE_c, NULL},
    {'f', IRCD_CMODE_f, get_flood},
    {'i', IRCD_CMODE_i, NULL},
    {'k', IRCD_CMODE_k, get_key},
    {'l', IRCD_CMODE_l, get_limit},
    {'m', IRCD_CMODE_m, NULL},
    {'n', IRCD_CMODE_n, NULL},
    {'p', IRCD_CMODE_p, NULL},
    {'r', IRCD_CMODE_r, NULL},
    {'s', IRCD_CMODE_s, NULL},
    {'t', IRCD_CMODE_t, NULL},
    {'u', IRCD_CMODE_u, NULL},
    {'z', IRCD_CMODE_z, NULL},
    {'A', IRCD_CMODE_A, NULL},
    {'C', IRCD_CMODE_C, NULL},
    {'G', IRCD_CMODE_G, NULL},
    {'H', IRCD_CMODE_H, NULL},
    {'K', IRCD_CMODE_K, NULL},
    {'L', IRCD_CMODE_L, get_redirect},
    {'N', IRCD_CMODE_N, NULL},
    {'O', IRCD_CMODE_O, NULL},
    {'Q', IRCD_CMODE_Q, NULL},
    {'R', IRCD_CMODE_R, NULL},
    {'S', IRCD_CMODE_S, NULL},
    {'V', IRCD_CMODE_V, NULL},
};

/* *INDENT-OFF* */
void moduleAddIRCDMsgs(void) {
    Message *m;

    m = createMessage("$",      denora_event_null); addCoreMessage(IRCD,m);		/* send routing table */
    m = createMessage("X",      denora_event_null); addCoreMessage(IRCD,m);		/* begin netburst NOW */
    m = createMessage("Y",      denora_event_eob); addCoreMessage(IRCD,m);		/* end of UPLINK netburst */
    m = createMessage("H",      denora_event_server); addCoreMessage(IRCD,m);		/* local sync start */
    m = createMessage("F",      denora_event_null); addCoreMessage(IRCD,m);		/* local sync end */
    m = createMessage("f",      denora_event_null); addCoreMessage(IRCD,m);		/* local U-lined sync end */
    m = createMessage("*",      denora_event_null); addCoreMessage(IRCD,m);		/* no operation */
    m = createMessage("|",      denora_event_null); addCoreMessage(IRCD,m);		/* oper type */
    m = createMessage("-",      denora_event_null); addCoreMessage(IRCD,m);		/* start mesh operation */
    m = createMessage("s",      denora_event_server); addCoreMessage(IRCD,m);	/* passive server response */
    m = createMessage("U",      denora_event_null); addCoreMessage(IRCD,m);		/* active service connect */
    m = createMessage("J",      denora_event_join); addCoreMessage(IRCD,m);		/* join */
    m = createMessage("k",      denora_event_kick); addCoreMessage(IRCD,m);		/* kick */
    m = createMessage("K",      denora_event_kill); addCoreMessage(IRCD,m);		/* kill */
    m = createMessage("m",      denora_event_mode); addCoreMessage(IRCD,m);		/* mode */
    m = createMessage("M",      denora_event_servermode); addCoreMessage(IRCD,m);		/* servermode */
    m = createMessage("N",      denora_event_nick); addCoreMessage(IRCD,m);		/* nick (introduce) */
    m = createMessage("n",      denora_event_nickchange); addCoreMessage(IRCD,m); /* nick (change) */
    m = createMessage("V",      denora_event_notice); addCoreMessage(IRCD,m);		/* notice */
    m = createMessage("L",      denora_event_part); addCoreMessage(IRCD,m);		/* part */
    m = createMessage("?",      denora_event_ping); addCoreMessage(IRCD,m);		/* ping */
    m = createMessage("P",      denora_event_privmsg); addCoreMessage(IRCD,m);		/* privmsg */
    m = createMessage("Q",      denora_event_quit); addCoreMessage(IRCD,m);		/* quit */
    m = createMessage("&",      denora_event_squit); addCoreMessage(IRCD,m);		/* squit */
    m = createMessage("t",      denora_event_topic); addCoreMessage(IRCD,m);		/* topic (user) */
    m = createMessage("T",      denora_event_servertopic); addCoreMessage(IRCD,m);	/* topic (server) */
    m = createMessage("@",      denora_event_null); addCoreMessage(IRCD,m);		/* wallops */
    m = createMessage("b",      denora_event_chghost); addCoreMessage(IRCD,m);		/* change displayed host */
    m = createMessage("a",      denora_event_chgname); addCoreMessage(IRCD,m);		/* change gecos */
    m = createMessage("!",      denora_event_pong); addCoreMessage(IRCD,m);    /* Pong */
    m = createMessage("E",      denora_event_error); addCoreMessage(IRCD,m);   /* Error */
    m = createMessage("~",      denora_event_null); addCoreMessage(IRCD,m);   /* undocumented */
    m = createMessage("_",      denora_event_null); addCoreMessage(IRCD,m);   /* JUPE */
    m = createMessage("v",      denora_event_version); addCoreMessage(IRCD,m);   /* JUPE */
}

/* *INDENT-ON* */

int denora_event_servermode(char *source, int ac, char **av)
{
    char* v[32];
    Channel* c;
    int i = 0;

    if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
    }

    if (*av[0] == '#') {
        c = findchan(av[0]);
        for (i = 1; i < ac; i++)
            v[i-1] = av[i];
        if (c) {
            chan_set_modes(c, ac-1, v);
            sql_do_chanmodes(c->name, v);
        }
    } else {
        v[0] = av[0];
        for (i = 0; i < ac; i++)
            v[i+1] = av[i];
        denora_event_mode(av[0], ac+1, v);
    }
    return MOD_CONT;
}

void inspircd_cmd_stats(char *sender, const char *letter, char *server)
{
    alog(LOG_PROTOCOL,"STATS sender %s, letter %s, server %s", sender, letter, server);
    /* currently not supported */
}

int denora_event_error(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    if (av[0]) {
        alog(LOG_ERROR, "ERROR: %s", av[0]);
    }
    return MOD_CONT;
}

/* since inspircd uses a random set of characters as the source we have
   to have this back together ourselves */
/* :oSmHcTj v inspircd.nomadirc.net InspIRCd-1.0(Beta6)+CVS Rev. 1.469 inspircd.nomadirc.net :Linux denora.elite-irc.net 2.6.9 (O=0) [SE=select] */
/*                    0               1                      2     3       4                   5 */
/* pFsqRvZ v denora.nomadirc.net Denora-1.0.1 (304) denora.nomadirc.net :InspIRCd 1.0 Beta 5 -- 1 */
/*                    0               1        2             3              4 */
int denora_event_version(char *source, int ac, char **av)
{
    char *v[32];
    char buf[BUFSIZE];
    *buf = '\0';

  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }

    if (ac == 6) {
      snprintf(buf, BUFSIZE - 1, "%s %s %s", av[1], av[2], av[3]);
    } else if (ac == 5) {
      snprintf(buf, BUFSIZE - 1, "%s %s", av[1], av[2]);
    } else {
      snprintf(buf, BUFSIZE - 1, "%s", av[1]);
    }
    v[0] = sstrdup(buf);
    sql_do_server_version(av[0], 1, v);
    denora_free(v[0]);
    return MOD_CONT;
}

int denora_event_notice(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    if (ac != 2) {
        return MOD_CONT;
    }
    m_notice(source, av[0], av[1]);
    return MOD_CONT;
}

void inspircd_cmd_version(char *server)
{
    send_cmd(s_StatServ, "VERSION %s", server);
}

void inspircd_cmd_motd(char *sender, char *server)
{
    send_cmd(sender, "MOTD %s", server);
}

int denora_event_nickchange(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    do_nick(av[0], av[1], NULL, NULL, NULL, NULL, 0, 0, 0, NULL, NULL, 0, NULL);
    return MOD_CONT;
}

void inspircd_cmd_nick(char *nick, char *name, const char *modes)
{
    /* N <time> <nick> <host> <displayed-host> <ident> <modes> <ipaddress> <server> :<GECOS> */
    send_cmd(CreateSum(), "N %lu %s %s %s %s %s 0.0.0.0 %s :%s",(long int) time(NULL),nick,ServiceHost,ServiceHost,ServiceUser,modes,ServerName,name);
}

void inspircd_cmd_mode(char *source, char *dest, char *buf)
{
    send_cmd(CreateSum(), "m %s %s %s", source, dest, buf);
}

void inspircd_cmd_bot_nick(char *nick, char *user, char *host, char *real,
                        char *modes)
{
    send_cmd(CreateSum(), "N %lu %s %s %s %s %s 0.0.0.0 %s :%s",(long int) time(NULL),nick,host,host,user,modes,ServerName,real);
}

void inspircd_cmd_notice(char *source, char *dest, char *buf)
{
    send_cmd(CreateSum(), "V %s %s :%s", source, dest, buf);
}

void inspircd_cmd_privmsg(char *source, char *dest, char *buf)
{
    send_cmd(CreateSum(), "P %s %s :%s", source, dest, buf);
}

void inspircd_cmd_serv_notice(char *source, char *dest, char *msg)
{
     if (!uplink) {
         send_cmd(CreateSum(), "V %s * :%s", source, msg); 
     } else if (!strcmp(dest,uplink)){ 
         alog(LOG_DEBUG, "debug: serv_notice output to %s", dest); 
         send_cmd(CreateSum(), "V %s * :%s", source, msg); 
    }
}

void inspircd_cmd_serv_privmsg(char *source, char *dest, char *msg)
{
 if (!uplink) {
         send_cmd(CreateSum(), "P %s * :%s", source, msg); 
 } else if (!strcmp(dest,uplink)){ 
         send_cmd(CreateSum(), "P %s * :%s", source, msg); 
 }
}

void inspircd_cmd_351(char *source)
{
    alog(LOG_PROTOCOL,"351 source %s", source);
}

/* QUIT */
void inspircd_cmd_quit(char *source, char *buf)
{
    if (buf) {
        send_cmd(CreateSum(), "Q %s :%s", source, buf);
    } else {
        send_cmd(CreateSum(), "QUIT %s :", source);
    }
}

/* PASS */
void inspircd_cmd_pass(char *pass)
{
    alog(LOG_PROTOCOL,"PASS pass %s", pass);
  /*   send_cmd(NULL, "PASS :%s", pass); */
}

/* SERVER name hop descript */
void inspircd_cmd_server(char *servname, int hop, char *descript)
{
    if (hop == 0)
    {
          send_cmd(CreateSum(), "H %s",servname);
    }
    else
    {
          send_cmd(CreateSum(), "U %s %s :%s", servname, RemotePassword, descript);
    }
}

/* PONG */
void inspircd_cmd_pong(char *servname, char *who)
{
    if (servname && who) {
        alog(LOG_PROTOCOL,"PONG: Server Name %s : Who %s", servname, who);
    }
    send_cmd(CreateSum(), "!");
}

/* JOIN */
void inspircd_cmd_join(char *user, char *channel, time_t chantime)
{
    alog(LOG_PROTOCOL,"Joining %s to %s at %ld", user, channel, (long int) chantime);
    send_cmd(CreateSum(), "J %s %s", user, channel);

    if (AutoOp)
        send_cmd(CreateSum(), "m %s %s +o %s", ServerName, channel, user);
}

/* PART */
void inspircd_cmd_part(char *nick, char *chan, char *buf)
{
    if (buf) {
        send_cmd(CreateSum(), "L %s %s :%s", nick, chan, buf);
    } else {
        send_cmd(CreateSum(), "L %s %s :", nick, chan);
    }
}

int denora_event_eob(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
  send_cmd(CreateSum(), "/ %s", s_StatServ);
  send_cmd(CreateSum(), "v %s Denora-%s %s :%s -- %s", ServerName, denora->version, 
                    ServerName, ircd->name, denora->build);
  return MOD_CONT;
}

/* GLOBOPS */
void inspircd_cmd_global(char *source, char *buf)
{
    send_cmd(CreateSum(), "V %s @* :%s", (source ? source : s_StatServ), buf);
}

/* SQUIT */
void inspircd_cmd_squit(char *servname, char *message)
{
    alog(LOG_PROTOCOL,"SQUIT servname %s, message %s", servname, message);
    send_cmd(CreateSum(), "& %s", servname);
}

void inspircd_cmd_connect()
{
    char *hops;

    srand(time(NULL));
    hops = sstrdup("0");
    me_server = do_server(NULL, ServerName, hops, ServerDesc, NULL);
    denora_free(hops);
    inspircd_cmd_server(ServerName, 1, ServerDesc);
}

/* Events */

int denora_event_ping(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    inspircd_cmd_pong(NULL, NULL);
    return MOD_CONT;
}

int denora_event_436(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    return MOD_CONT;
}

int denora_event_pong(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
  return MOD_CONT;
}

int denora_event_away(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    return MOD_CONT;
}

/* T 1114880269 Trystan #ircops :test tes test */
/*    0          1       2        3 */
int denora_event_servertopic(char *source, int ac, char **av)
{
    char *v[32];

  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }

    if (ac != 4) {
        return MOD_CONT;
    }

    v[0] = av[2];  /* channel */
    v[1] = av[1];  /* setter  */
    v[2] = av[0];  /* timestamp */
    v[3] = av[3];  /* topic */

    do_topic(4, v); /* no source (server set) */
    return MOD_CONT;
}

int denora_event_topic(char *source, int ac, char **av)
{
    char *v[32];

  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }

    if (ac != 3) {
        return MOD_CONT;
    }
    v[0] = av[1]; /* channel */
    v[1] = av[0]; /* setter */
    v[2] = av[2]; /* topic content */
    do_topic(3, v);
    return MOD_CONT;
}

int denora_event_squit(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    if (ac != 1)
        return MOD_CONT;
    do_squit(av[1]);
    return MOD_CONT;
}

int denora_event_quit(char *source, int ac, char **av)
{
     char *v[32];
     if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
     }
    if (ac != 2)
        return MOD_CONT;
    v[0] = av[1];
    v[1] = av[2];
    do_quit(av[0], 2, v);
    return MOD_CONT;
}

int denora_event_mode(char *source, int ac, char **av)
{
    char *v[32];
    int i;

  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }

    if (ac < 3)
        return MOD_CONT;

    for (i = 1; i < ac; i++)
    {
        v[i-1] = av[i];
    }
    if (*av[1] == '#') {
        do_cmode(av[1], ac-1, v);
    } else {
        do_umode(av[0], ac-1, v);
    }
    return MOD_CONT;
}

int denora_event_kill(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    if (ac != 3)
        return MOD_CONT;
    m_kill(source, av[0], av[1]);
    return MOD_CONT;
}

/* Set mod_current_buffer from here */ 
void inspircd_set_mod_current_buffer(int ac, char **av) 
{ 
    if (ac >= 3) {
        mod_current_buffer = sstrdup(av[2]); 
    } else {
        mod_current_buffer = NULL; 
    }
} 

int denora_event_kick(char *source, int ac, char **av)
{
    char *v[32];

  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }

    /* Syntax: k <source> <dest> <channel> :<reason> */
    if (ac != 4)
        return MOD_CONT;
    v[0] = av[1];
    v[1] = av[2];
    v[2] = av[3];
    do_kick(av[0], 3, v);
    return MOD_CONT;
}

int denora_event_join(char *source, int ac, char **av)
{
    User *u;
    Channel *c;
    int i;
    char* new_av[32];
    char* joinumodes[3];
    char thismodes[256];
    joinumodes[0] = thismodes;

  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    /* some sanity checks */
    if (ac < 2) {
            return MOD_CONT;
    }

    u = finduser(av[0]); /*  get the user info */

    if (!u) {
        return MOD_CONT;
    }

    joinumodes[1] = av[0];

    /* iterate through each channel the user is joining */
    for (i = 1; i < ac; i++)
    {
            /* the channel name */
            new_av[0] = av[i];

            /* strip off the permissions */
            strcpy(thismodes,"");
            if (*new_av[0] != '#')
            {
		    switch (*new_av[0])
		    {
			case '@':
				strcpy(thismodes,"+o");
			break;
			case '%':
				strcpy(thismodes,"+h");
			break;
			case '+':
				strcpy(thismodes,"+v");
			break;
		    }
                    new_av[0]++;
            }

            do_join(av[0], 1, new_av);
            c = findchan(new_av[0]);
            if ((c) && (strcmp(thismodes,"")))
            {
                    chan_set_modes(c, 2, joinumodes);
                    sql_do_chanmodes(c->name, joinumodes);
            }
    }
    return MOD_CONT;
}

int denora_event_motd(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
  return MOD_CONT;
}

int denora_event_setname(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    change_user_realname(av[0], av[1]);
    return MOD_CONT;
}

int denora_event_chgname(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    change_user_realname(av[0], av[1]);
    return MOD_CONT;
}

int denora_event_setident(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    return MOD_CONT;
}

int denora_event_chgident(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    return MOD_CONT;
}

int denora_event_sethost(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    change_user_host(av[0], av[1]);
    return MOD_CONT;
}

int denora_event_nick(char *source, int ac, char **av)
{
    User *user;
    struct in_addr addy;
    uint32* ad = (uint32*)&addy;

  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }

    /* Syntax: N <time> <nick> <host> <displayed-host> <ident> <modes> <ipaddress> <server> :<GECOS> */

    if (ac == 9) {
         inet_aton(av[6],&addy);
         *source = '\0';
         user = do_nick(source, av[1], av[4], av[2], av[7], av[8],
               strtoul(av[0], NULL, 10), 0, htonl(*ad), av[3], NULL, 0, av[5]);
         if (user)
            denora_set_umode(user, 1, &av[5]);
    }
    return MOD_CONT;
}

int denora_event_chghost(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    change_user_host(av[0], av[1]);
    return MOD_CONT;
}

/* EVENT: SERVER */
int denora_event_server(char *source, int ac, char **av)
{
    char *hops;

   /*  s inspircd.nomadirc.net linkpass 7000 :Waddle World */
    /* do_server(const char *source, char *servername, char *hops,
                  char *descript, char *numeric) */
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }

    /* since the source is a random set of numbers 
       we need to reset it */
    *source = '\0';

    if (ac == 1) {
      hops = sstrdup("2");
      do_server(denora->uplink, av[0], hops, NULL, NULL);

    } else {
        hops = sstrdup("1");
        if (!denora->uplink) {
	        denora->uplink = sstrdup(av[0]);
            me_server->uplink = sstrdup(av[0]);
            do_server(source, av[0], hops, av[3], NULL);
        }
    }
    denora_free(hops);
    return MOD_CONT;
}

int denora_event_privmsg(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    if (ac != 3)
        return MOD_CONT;
    m_privmsg(av[0], av[1], av[2]);
    return MOD_CONT;
}

int denora_event_part(char *source, int ac, char **av)
{
    char *v[32];

  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    if (ac != 3)
        return MOD_CONT;
    v[0] = av[1]; /* channel */
    v[1] = av[2]; /* reason */
    do_part(av[0], 2, v);
    return MOD_CONT;
}

int denora_event_whois(char *source, int ac, char **av)
{
  if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
  }
    return MOD_CONT;
}

void inspircd_cmd_eob()
{
    send_cmd(CreateSum(), "F %ld", (long int) time(NULL));
}

void inspircd_cmd_ping(char *server)
{
    alog(LOG_PROTOCOL,"PING server %s", server);
}

void inspircd_cmd_ctcp(char *source, char *dest, char *buf)
{
   send_cmd(CreateSum(), "V %s %s :\1%s \1", source, dest, buf);
}

void moduleAddIRCDCmds() 
{
    pmodule_set_mod_current_buffer(inspircd_set_mod_current_buffer);
    pmodule_cmd_nick(inspircd_cmd_nick);
    pmodule_cmd_mode(inspircd_cmd_mode);
    pmodule_cmd_bot_nick(inspircd_cmd_bot_nick);
    pmodule_cmd_notice(inspircd_cmd_notice);
    pmodule_cmd_privmsg(inspircd_cmd_privmsg);
    pmodule_cmd_serv_notice(inspircd_cmd_serv_notice);
    pmodule_cmd_serv_privmsg(inspircd_cmd_serv_privmsg);
    pmodule_cmd_351(inspircd_cmd_351);
    pmodule_cmd_quit(inspircd_cmd_quit);
    pmodule_cmd_pong(inspircd_cmd_pong);
    pmodule_cmd_join(inspircd_cmd_join);
    pmodule_cmd_part(inspircd_cmd_part);
    pmodule_cmd_global(inspircd_cmd_global);
    pmodule_cmd_squit(inspircd_cmd_squit);
    pmodule_cmd_connect(inspircd_cmd_connect);
    pmodule_cmd_eob(inspircd_cmd_eob);
    pmodule_cmd_ctcp(inspircd_cmd_ctcp);
    pmodule_cmd_version(inspircd_cmd_version);
    pmodule_cmd_stats(inspircd_cmd_stats);
    pmodule_cmd_motd(inspircd_cmd_motd);
    pmodule_cmd_ping(inspircd_cmd_ping);
}

int DenoraInit(int argc, char **argv) 
{
  if (denora->protocoldebug) {
        protocol_debug(NULL, argc, argv);
  }

    moduleAddAuthor("Denora");
    moduleAddVersion("$Id: inspircd.c 433 2005-07-26 16:58:24Z drstein $");
    moduleSetType(PROTOCOL);

    UseTokens = 1;

    pmodule_ircd_version("InspIRCd 1.0 Beta5");
    pmodule_ircd_cap(myIrcdcap);
    pmodule_ircd_var(myIrcd);
    pmodule_ircd_cbmodeinfos(myCbmodeinfos);
    pmodule_ircd_cbmodes(myCbmodes);
    pmodule_ircd_cmmodes(myCmmodes);
    pmodule_ircd_umodes(myumodes);
    pmodule_ircd_csmodes(myCsmodes);
    pmodule_ircd_useTSMode(0);
    pmodule_irc_var(IRC_INSPIRCD);
    pmodule_oper_umode(IRC_UMODE_o);
    moduleAddIRCDCmds();
    moduleAddIRCDMsgs();
    return MOD_CONT;
}

