
/*
 *
 * (C) 2004-2005 Denora Team
 * Contact us at info@nomadirc.net
 *
 * Please read COPYING and README for furhter details.
 *
 * Based on the original code of Anope by Anope Team.
 * Based on the original code of Thales by Lucas.
 * 
 * $Id: rdb.c 421 2005-07-17 09:03:20Z trystan $
 *
 */
#include "denora.h"

char *rdb_errmsg;

/*************************************************************************/

int rdb_init()
{
    int res = 0;
#ifdef USE_MYSQL
    if (sqltype == SQL_MYSQL) {
        res = db_mysql_init();
    }
	SET_SEGV_LOCATION();
#endif
#ifdef USE_POSTGRE
    if (sqltype == SQL_POSTGRE) {
        res = db_postgre_init();
    }
	SET_SEGV_LOCATION();
#endif
    if (res) {
        db_connect();
    }
    return res;
}

/*************************************************************************/

int rdb_close()
{
#ifdef USE_MYSQL
    if (sqltype == SQL_MYSQL) {
     if (mysql) {
        mysql_close(mysql);
      }
      mysql = NULL;
      return 1;
    }
#endif
#ifdef USE_POSTGRE
    if (sqltype == SQL_POSTGRE) {
        db_postgre_close();
        return 1;
    }
#endif
    return 0;
}

/*************************************************************************/

int rdb_clear_table(char *table)
{
    static char buf[1024];

    SET_SEGV_LOCATION();

#ifdef USE_MYSQL
    if (sqltype == SQL_MYSQL) {
        snprintf(buf, sizeof(buf), "TRUNCATE TABLE %s", table);
        return db_mysql_query(buf);
    }
#endif
#ifdef USE_POSTGRE
    if (sqltype == SQL_POSTGRE) {
        snprintf(buf, sizeof(buf), "TRUNCATE TABLE %s", table);
        return db_postgre_query(buf);        
    }
#endif
    return 0;
}

/*************************************************************************/

int rdb_direct_query(char *query)
{
    SET_SEGV_LOCATION();

    if (!denora->do_sql) {
        return -1;
    }

#ifdef USE_MYSQL
    if (sqltype == SQL_MYSQL) {
        return db_mysql_query(query);
    }
#endif
#ifdef USE_POSTGRE
    if (sqltype == SQL_POSTGRE) {    
        return db_postgre_query(query);
    }
#endif
    return 0;
}

/*************************************************************************/

/* send an SQL query to the database */
int rdb_query(const char *fmt, ...)
{
    va_list args;
    char buf[NET_BUFSIZE];             /* should be enough for long queries (don't underestimate :o)) */

    if (!denora->do_sql) {
        return -1;
    }

    if (BadPtr(fmt)) {
        return -1;
    }

    va_start(args, fmt);
    vsprintf(buf, fmt, args);
    rdb_direct_query(buf);
    va_end(args);
    return 0;
}

/*************************************************************************/

/* escape the string */
char *rdb_escape(char *ch)
{
#ifdef USE_MYSQL
    if (sqltype == SQL_MYSQL) {
        return db_mysql_quote(ch);
    }
#endif
#ifdef USE_POSTGRE
    if (sqltype == SQL_POSTGRE) {
        return db_postgre_quote(ch);
    }
#endif
    return NULL;
}

/*************************************************************************/

int rdb_insertid()
{
    SET_SEGV_LOCATION();

#ifdef USE_MYSQL
    if (sqltype == SQL_MYSQL) {
        return mysql_insert_id(mysql);
    }
#endif
#ifdef USE_POSTGRE
    if (sqltype == SQL_POSTGRE) {
        return 0;
    }
#endif 
    return 0;
}

/*************************************************************************/

char *rdb_error_msg()
{
    denora_free(rdb_errmsg);
    rdb_errmsg = NULL;
#ifdef USE_MYSQL
    if (sqltype == SQL_MYSQL) {
        rdb_errmsg = sstrdup(mysql_error(mysql));
        return rdb_errmsg;
    }
#endif
#ifdef USE_POSTGRE
    if (sqltype == SQL_POSTGRE) {
        rdb_errmsg = sstrdup(PQerrorMessage(postgre));
    }
#endif
    return rdb_errmsg;
}

