
/*
 *
 * (C) 2004-2005 Denora Team
 * Contact us at info@nomadirc.net
 *
 * Please read COPYING and README for furhter details.
 *
 * Based on the original code of Anope by Anope Team.
 * Based on the original code of Thales by Lucas.
 * 
 * $Id: tld.c 391 2005-07-06 03:28:36Z trystan $
 *
 */

#include "denora.h"

void insert_tld(TLD * t);

TLD *tldlist[1024];
TLD *tldlisttemp[1024];

static TLD *current;
static int next_index;

void sql_do_tld(int type, char *code, char *country, int count,
                int overall);
void delete_tld(TLD * t);

/*************************************************************************/

void tld_update(char *country_code)
{
    TLD *t;

    t = findtld(country_code);
    if (t) {
        t->count--;
    }
    rdb_query("UPDATE %s SET count=%d, overall=%d WHERE code=\'%s\'",
              TLDTable, t->count, t->overall, country_code);
}

/* NICK (new nick) */
void sql_do_tld(int type, char *code, char *country, int count,
                int overall)
{
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    if (!denora->do_sql) {
        return;
    }
    SET_SEGV_LOCATION();
    if (type == 1) {
        rdb_query("SELECT country FROM %s WHERE country = \"%s\";",
                  TLDTable, country);
#ifdef USE_MYSQL
        mysql_res = mysql_store_result(mysql);
        if (mysql_res) {
            if (mysql_num_rows(mysql_res) == 0) {       /* exists this nick already in the database? */
                rdb_query
                    ("UPDATE %s SET count=%d, overall=%d WHERE code=\'%s\'",
                     TLDTable, count, overall, code);
            } else {
                rdb_query
                    ("INSERT INTO %s (code, country, count, overall) VALUES(\'%s\', \'%s\', %d, %d)",
                     TLDTable, code, country, count, overall);
            }
            mysql_free_result(mysql_res);
        }
#endif
    }
    if (type == 4) {
        rdb_query("SELECT country FROM %s WHERE country = \"%s\";",
                  TLDTable, country);
#ifdef USE_MYSQL
        mysql_res = mysql_store_result(mysql);
        if (mysql_res) {
            if (mysql_num_rows(mysql_res) == 0) {       /* exists this nick already in the database? */
                rdb_query
                    ("INSERT INTO %s (code, country, count, overall) VALUES(\'%s\', \'%s\', %d, %d)",
                     TLDTable, code, country, count, overall);
            } else {
                rdb_query
                    ("UPDATE %s SET count=%d, overall=%d WHERE code=\'%s\'",
                     TLDTable, count, overall, code);
            }
            mysql_free_result(mysql_res);
        }
#endif
    }
}

/*************************************************************************/

static TLD *new_tld(const char *country)
{
    TLD *t, **list;

    SET_SEGV_LOCATION();

    t = scalloc(sizeof(TLD), 1);
    if (!country) {
        country = sstrdup("??");
        alog(LOG_DEBUG, "debug: country is NULL set it to %s", country);
    }

    strscpy(t->countrycode, country, NICKMAX);
    list = &tldlist[TLDHASH(t->countrycode)];
    t->next = *list;
    if (*list)
        (*list)->prev = t;
    *list = t;
    return t;
}

/*************************************************************************/

void load_tld_db(void)
{
    DenoraDBFile *dbptr = scalloc(1, sizeof(DenoraDBFile));
    TLD *t = NULL;
    char *key, *value;
    int retval = 0;

    fill_db_ptr(dbptr, 0, TLD_VERSION, s_StatServ, TLDDB);

    /* let's remove existing temp files here, because we only load dbs on startup */
    remove(dbptr->temp_name);

    /* Open the db, fill the rest of dbptr and allocate memory for key and value */
    if (new_open_db_read(dbptr, &key, &value)) {
        if (dbptr->fptr) {
            fclose(dbptr->fptr);
        }
        free(dbptr);
        return;                 /* Bang, an error occurred */
    }

    while (1) {
        /* read a new entry and fill key and value with it -Certus */
        retval = new_read_db_entry(&key, &value, dbptr->fptr);

        if (retval == DB_READ_ERROR) {
            alog(LOG_NORMAL, "WARNING! DB_READ_ERROR in %s",
                 dbptr->filename);
            new_close_db(dbptr->fptr, &key, &value);
            free(dbptr);
            return;
        } else if (retval == DB_EOF_ERROR) {
            alog(LOG_EXTRADEBUG, "debug: %s read successfully",
                 dbptr->filename);
            new_close_db(dbptr->fptr, &key, &value);
            free(dbptr);
            return;
        } else if (retval == DB_READ_BLOCKEND) {        /* DB_READ_BLOCKEND */
            /* a ctcp entry has completely been read. put any checks in here! */
        } else {
            /* DB_READ_SUCCESS */

            if (!*value || !*key)
                continue;

            if (!stricmp(key, "ccode")) {
                t = new_tld(value);
            } else if (!stricmp(key, "country")) {
                t->country = sstrdup(value);
            } else if (!stricmp(key, "overall")) {
                t->overall = atoi(value);
            }
        }                       /* else */
    }                           /* while */
}

/*************************************************************************/

void save_tld_db(void)
{
    DenoraDBFile *dbptr = scalloc(1, sizeof(DenoraDBFile));
    TLD *t;
    int i;

    fill_db_ptr(dbptr, 0, TLD_VERSION, s_StatServ, TLDDB);

    /* time to backup the old db */
    rename(TLDDB, dbptr->temp_name);

    if (new_open_db_write(dbptr)) {
        rename(dbptr->temp_name, TLDDB);
        free(dbptr);
        return;                 /* Bang, an error occurred */
    }

    for (i = 0; i < 1024; i++) {
        for (t = tldlist[i]; t; t = t->next) {
            new_write_db_entry("ccode", dbptr, "%s", t->countrycode);
            new_write_db_entry("country", dbptr, "%s", t->country);
            new_write_db_entry("overall", dbptr, "%d", t->overall);
            new_write_db_endofblock(dbptr);
        }
    }

    if (dbptr) {
        new_close_db(dbptr->fptr, NULL, NULL);  /* close file */
        remove(dbptr->temp_name);       /* saved successfully, no need to keep the old one */
        free(dbptr);            /* free the db struct */
    }
}

/*************************************************************************/

void delete_tld(TLD * t)
{
    SET_SEGV_LOCATION();

    alog(LOG_EXTRADEBUG, "debug: delete_tld() called");

    denora_free(t->country);

    alog(LOG_EXTRADEBUG, "debug: delete_tld(): delete from list");

    SET_SEGV_LOCATION();
    if (t->prev) {
        t->prev->next = t->next;
    } else {
        tldlist[TLDHASH(t->country)] = t->next;
    }
    if (t->next) {
        t->next->prev = t->prev;
    }
    SET_SEGV_LOCATION();

    alog(LOG_EXTRADEBUG, "debug: delete_tld(): free user structure");

    denora_free(t);
    alog(LOG_EXTRADEBUG, "debug: delete_tld() done");
}

/*************************************************************************/

TLD *findtld(const char *country)
{
    TLD *t;

    SET_SEGV_LOCATION();

    if (!*country) {
        alog(LOG_DEBUG, "Error: findtld() called with NULL values");
        do_backtrace(1);
        return NULL;
    }

    alog(LOG_EXTRADEBUG, "debug: findtld(%p)", country);
    t = tldlist[TLDHASH(country)];
    while (t && stricmp(t->countrycode, country) != 0)
        t = t->next;
    alog(LOG_EXTRADEBUG, "debug: findtld(%s) -> 0x%p", country,
         (void *) t);
    return t;
}

/*************************************************************************/

TLD *do_tld(char *country, char *code)
{
    TLD *t, *t2;

    if (code) {
        t = findtld(code);
    } else {
        code = sstrdup("??");
        t = findtld(code);
    }

    SET_SEGV_LOCATION();

    if (!t) {

        /* Allocate User structure and fill it in. */
        t2 = new_tld(code);
        if (country) {
            t2->country = sstrdup(country);
        } else {
            t2->country = sstrdup("Unknown");
        }
        t2->count = 1;
        t2->overall = 1;
        sql_do_tld(ADD, code, t2->country, t2->count, t2->overall);
        return t2;

    } else {
        t->count++;
        if (t->count > t->overall) {
            t->overall = t->count;
        }
        if (!t->overall) {
            t->overall = 1;
        }
        sql_do_tld(UPDATE, code, t->country, t->count, t->overall);
        return t;
    }
}

/*************************************************************************/

TLD *first_tld(void)
{
    next_index = 0;

    SET_SEGV_LOCATION();

    while (next_index < 1024 && current == NULL)
        current = tldlist[next_index++];
    alog(LOG_EXTRADEBUG, "debug: first_tld() returning %s",
         current ? current->countrycode : "NULL (end of list)");
    return current;
}

/*************************************************************************/

TLD *next_tld(void)
{
    SET_SEGV_LOCATION();

    if (current)
        current = current->next;
    if (!current && next_index < 1024) {
        while (next_index < 1024 && current == NULL)
            current = tldlist[next_index++];
    }
    alog(LOG_EXTRADEBUG, "debug: next_tld() returning %s",
         current ? current->countrycode : "NULL (end of list)");
    return current;
}

/*************************************************************************/


